# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Modules that implement mappings between pytrees.

Pytree transforms describe a broad collection of mappings between pytrees. Most
of these can be crudely separated into two categories:
  1. Operations that transform individual arrays (subset of) within a pytree.
     [e.g. rescaling, reshaping, broadcasting, or other transforms]
  2. Operations that return elements (static) or a result of the computation
     dependent on inputs to the output pytree.
     [e.g. feature injection, input featurization, variable transformation]
"""

from __future__ import annotations

import abc
import dataclasses
import functools
import re
from typing import Callable, Protocol, Sequence

from dinosaur import radiation
from dinosaur import spherical_harmonic
from flax import nnx
import jax
import jax.numpy as jnp
from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import data_specs
from neuralgcm.experimental import dynamic_io
from neuralgcm.experimental import orographies
from neuralgcm.experimental import pytree_utils
from neuralgcm.experimental import random_processes
from neuralgcm.experimental import spatial_filters
from neuralgcm.experimental import typing
from neuralgcm.experimental import units
from neuralgcm.experimental import xarray_utils
import numpy as np
import xarray


ShapeFloatStruct = typing.ShapeFloatStruct


class TransformParams(nnx.Variable):
  """Custom variable class for transform parameters."""


class Transform(Protocol):
  """Protocol for pytree transforms."""

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    ...

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    ...


TransformFactory = Callable[..., Transform]


class TransformABC(nnx.Module, abc.ABC):
  """Abstract base class for pytree transforms."""

  @abc.abstractmethod
  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    raise NotImplementedError()

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    return nnx.eval_shape(self.__call__, input_shapes)


class Broadcast(TransformABC):
  """Features transform that broadcasts all features."""

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    leaves, tree_def = jax.tree.flatten(inputs)
    leaves = jnp.broadcast_arrays(*leaves)
    return jax.tree.unflatten(tree_def, leaves)


class Identity(TransformABC):
  """Transform does not modify inputs."""

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    return inputs


class Empty(TransformABC):
  """Transform that returns an empty dict."""

  def __call__(self, inputs: ...) -> typing.Pytree:
    return {}


@dataclasses.dataclass
class NestDict(TransformABC):
  """Transform that nests elements of a flat dictionary into a nested dict."""

  keys_to_nest: tuple[str, ...]
  nested_key_name: str

  def __call__(self, inputs: dict[str, typing.Array]) -> typing.Pytree:
    nested = {k: v for k, v in inputs.items() if k in self.keys_to_nest}
    non_nested = {k: v for k, v in inputs.items() if k not in self.keys_to_nest}
    return non_nested | {self.nested_key_name: nested}


# TODO(dkochkov): Consider supporting use_bias use_scale selectively.
# could use dict[str, bool] variation, or even default_dict if fiddle is happy.
class BatchShiftAndNormalize(TransformABC):
  """Shift & norm over feature axis with support for batch stat updates."""

  def __init__(
      self,
      feature_sizes: dict[str, int],
      feature_axis: int = -3,
      use_running_average: bool = True,
      use_bias: bool = False,
      use_scale: bool = False,
      momentum: float = 0.99,
      epsilon: float = 1e-5,
      bias_init: nnx.initializers.Initializer = nnx.initializers.zeros_init(),
      scale_init: nnx.initializers.Initializer = nnx.initializers.ones_init(),
      batch_axis_name: str | None = None,
      skip_unspecified: bool = False,
      rngs: nnx.Rngs | None = None,
  ):
    self.batch_transforms = {
        k: nnx.BatchNorm(
            num_features=v,
            use_running_average=False,  # controlled by class attribute.
            axis=feature_axis,
            momentum=momentum,
            epsilon=epsilon,
            use_scale=use_scale,
            use_bias=use_bias,
            bias_init=bias_init,
            scale_init=scale_init,
            axis_name=batch_axis_name,
            rngs=rngs,
        )
        for k, v in feature_sizes.items()
    }
    self.skip_unspecified = skip_unspecified
    self.use_running_average = use_running_average

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    transforms = pytree_utils.replace_with_matching_or_default(
        inputs,
        self.batch_transforms,
        default=lambda x, _: x if self.skip_unspecified else None,
        check_used_all_replace_keys=False,
    )
    use_avg = self.use_running_average
    outputs = jax.tree.map(lambda x, y: y(x, use_avg), inputs, transforms)
    return from_dict_fn(outputs)

  @classmethod
  def for_input_shapes(
      cls,
      input_shapes: typing.Pytree,
      feature_axis: int = -3,
      use_running_average: bool = True,
      use_bias: bool = False,
      use_scale: bool = False,
      momentum: float = 0.99,
      epsilon: float = 1e-5,
      bias_init: nnx.initializers.Initializer = nnx.initializers.zeros_init(),
      scale_init: nnx.initializers.Initializer = nnx.initializers.ones_init(),
      batch_axis_name: str | None = None,
      skip_unspecified: bool = False,
      rngs: nnx.Rngs | None = None,
  ):
    """Custom constructor based on input shapes that should be normalized."""
    feature_sizes = jax.tree.map(lambda x: x.shape[feature_axis], input_shapes)
    return cls(
        feature_sizes=feature_sizes,
        feature_axis=feature_axis,
        use_running_average=use_running_average,
        use_bias=use_bias,
        use_scale=use_scale,
        momentum=momentum,
        epsilon=epsilon,
        bias_init=bias_init,
        scale_init=scale_init,
        batch_axis_name=batch_axis_name,
        skip_unspecified=skip_unspecified,
        rngs=rngs,
    )


@dataclasses.dataclass
class FeatureSelector(TransformABC):
  """Features transform that retains items whose keys match against regex.

  Attributes:
    regex_patterns: regular expression pattern that specifies the set of keys
      from `inputs` that will be returned by __call__ method.
  """

  regex_patterns: str

  def __call__(
      self,
      inputs: dict[str, typing.Array],
  ) -> dict[str, typing.Array]:
    outputs = {}
    for k, v in inputs.items():
      if re.fullmatch(self.regex_patterns, k):
        outputs[k] = v
    return outputs


class ComposedTransform(nnx.Module):
  """Transform that composes a sequence of transforms."""

  def __init__(
      self,
      transforms: Sequence[Transform],
  ):
    self.transforms = transforms

  def __call__(self, inputs):
    for transform in self.transforms:
      inputs = transform(inputs)
    return inputs

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    # We use explicit implementation here to allow subcomponents to call custom
    # output_shapes methods.
    for transform in self.transforms:
      input_shapes = transform.output_shapes(input_shapes)
    return input_shapes


# TODO(dkochkov): Think whether these can have default init + init from file.
class ShiftAndNormalize(TransformABC):
  """Transforms inputs by shifting and normalizing values by `shifts/scales`."""

  def __init__(
      self,
      shifts: typing.Pytree,
      scales: typing.Pytree,
      global_scale: float | None = None,
  ):
    if global_scale is not None:
      scales = jax.tree_util.tree_map(lambda x: x * global_scale, scales)
    self.shifts = TransformParams(shifts)
    self.scales = TransformParams(scales)

  def update_shifts_and_scales(self, shifts, scales):
    """Updates values of `shifts` and `scales`."""
    self.shifts = TransformParams(shifts)
    self.scales = TransformParams(scales)

  def __call__(self, inputs):
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    shifts = pytree_utils.replace_with_matching_or_default(
        inputs,
        self.shifts.value,
        default=None,
        check_used_all_replace_keys=False,
    )
    scales = pytree_utils.replace_with_matching_or_default(
        inputs,
        self.scales.value,
        default=None,
        check_used_all_replace_keys=False,
    )
    # if shifts/scales have missing values present in `inputs`, we insert `None`
    # for the default. If corresponding `inputs` is not `None`, this will raise
    # an error, as expected. This works because tree_map skips `None` values in
    # the first argument, as long as all dictionary keys match.
    result = jax.tree_util.tree_map(
        lambda x, y, z: (x - y) / z, inputs, shifts, scales
    )
    return from_dict_fn(result)

  @classmethod
  def identity_init(cls, input_shapes: typing.Pytree):
    """Initializes the module with identity transform."""
    shifts = jax.tree.map(jnp.zeros, input_shapes)
    scales = jax.tree.map(jnp.ones, input_shapes)
    return cls(shifts=shifts, scales=scales)


class InverseShiftAndNormalize(TransformABC):
  """Inverse of the `ShiftAndNormalize` for the same `shifts/scales`."""

  def __init__(
      self,
      shifts: typing.Pytree,
      scales: typing.Pytree,
      global_scale: float | None = None,
  ):
    if global_scale is not None:
      scales = jax.tree_util.tree_map(lambda x: x * global_scale, scales)
    self.shifts = TransformParams(shifts)
    self.scales = TransformParams(scales)

  def update_shifts_and_scales(self, shifts, scales):
    """Updates values of `shifts` and `scales`."""
    self.shifts = TransformParams(shifts)
    self.scales = TransformParams(scales)

  def __call__(self, inputs):
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    shifts = pytree_utils.replace_with_matching_or_default(
        inputs,
        self.shifts.value,
        default=None,
        check_used_all_replace_keys=False,
    )
    scales = pytree_utils.replace_with_matching_or_default(
        inputs,
        self.scales.value,
        default=None,
        check_used_all_replace_keys=False,
    )
    # if shifts/scales have missing values present in `inputs`, we insert `None`
    # for the default. If corresponding `inputs` is not `None`, this will raise
    # an error, as expected. This works because tree_map skips `None` values in
    # the first argument, as long as all dictionary keys match.
    result = jax.tree_util.tree_map(
        lambda x, y, z: x * z + y, inputs, shifts, scales
    )
    return from_dict_fn(result)

  @classmethod
  def identity_init(cls, input_shapes: typing.Pytree):
    """Initializes the module with identity transform."""
    shifts = jax.tree.map(jnp.zeros, input_shapes)
    scales = jax.tree.map(jnp.ones, input_shapes)
    return cls(shifts=shifts, scales=scales)


class TakeSurfaceAdjacentSgimaLevel(TransformABC):
  """Takes surface adjacent sigma level from the input."""

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    """Returns `inputs` where only last sigma level is retained."""

    def _slice_fn(x):
      return x[slice(-1, None), ...]

    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    outputs = jax.tree.map(_slice_fn, inputs)
    return from_dict_fn(outputs)


@dataclasses.dataclass
class ClipWavenumbers(TransformABC):
  """Sets top `wavenumbers_to_clip` total wavenumbers to zero in the input."""

  grid: coordinates.SphericalHarmonicGrid
  wavenumbers_to_clip: int = 1

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    """Returns `inputs` where only last sigma level is retained."""
    return self.grid.ylm_grid.clip_wavenumbers(inputs, self.wavenumbers_to_clip)


class MaskTransform(TransformABC):
  """Masks arrays in the input dictionary with keys in `fields_to_mask`."""

  def __init__(
      self,
      fields_to_mask: Sequence[str],
      mask_shape: tuple[int, ...],
      fill_value_true: float = 0.0,
      fill_value_false: float = 1.0,
      *,
      fill_threshold: float = jnp.nan,
  ):
    """Fills `fill_value` in `fields_to_mask` entries of an input pytree."""
    self.mask_shape = mask_shape
    self.mask = TransformParams(jnp.zeros(mask_shape))
    self.fill_value_true = fill_value_true
    self.fill_value_false = fill_value_false
    self.fill_threshold = fill_threshold
    self.fields_to_mask = fields_to_mask

  def update_from_xarray(
      self, dataset: xarray.Dataset, threshold_variable: str
  ):
    dataset = dataset[threshold_variable]
    if threshold_variable == 'sea_ice_cover':
      # TODO(asubel): This is a hack, fix this.
      dataset = dataset.max('time').squeeze()
    if 'time' in dataset.dims:
      dataset = dataset.isel(time=0).squeeze()
    if dataset.shape != self.mask_shape:
      raise ValueError(
          f'dataset.shape={dataset.shape=} does not match'
          f' self.mask.shape={self.mask_shape=}'
      )
    if self.fill_threshold is jnp.nan:
      mask_value = xarray.where(
          np.isnan(dataset), self.fill_value_true, self.fill_value_false
      )
    else:
      mask_value = xarray.where(
          dataset > self.fill_threshold,
          self.fill_value_true,
          self.fill_value_false,
      )
    self.mask = TransformParams(jnp.array(mask_value.values))

  def update_from_pytree(self, data):
    assert set(data.keys()) == set(self.mask.keys())
    self.mask = TransformParams(data)

  def update_from_array(self, mask: typing.Array):
    if mask.shape != self.mask_shape:
      raise ValueError(
          f'mask.shape={mask.shape=} does not match'
          f' self.mask.shape={self.mask.shape=}'
      )
    self.mask = TransformParams(mask)

  def __call__(self, inputs):
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    for k, v in inputs.items():
      if k in self.fields_to_mask:
        inputs[k] = v * self.mask.value
    return from_dict_fn(inputs)


@dataclasses.dataclass
class ToModalWithDivCurl(TransformABC):
  """Module that converts inputs to modal replacing velocity with div/curl."""

  grid: coordinates.SphericalHarmonicGrid | coordinates.LonLatGrid
  u_key: str = 'u_component_of_wind'
  v_key: str = 'v_component_of_wind'

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    dinosaur_grid = self.grid.ylm_grid
    if self.u_key not in inputs or self.v_key not in inputs:
      raise ValueError(
          f'{(self.u_key, self.v_key)=} not found in {inputs.keys()=}'
      )
    sec_lat = 1 / dinosaur_grid.cos_lat
    u, v = inputs.pop(self.u_key), inputs.pop(self.v_key)
    # here u,v stand for velocity / cos(lat), but the cos(lat) is cancelled in
    # divergence and curl operators below.
    inputs[self.u_key] = u * sec_lat
    inputs[self.v_key] = v * sec_lat
    to_modal_fn = lambda x: dinosaur_grid.to_modal(x) if x is not None else None
    modal_outputs = jax.tree_util.tree_map(to_modal_fn, inputs)
    u, v = modal_outputs.pop(self.u_key), modal_outputs.pop(self.v_key)
    modal_outputs['divergence'] = dinosaur_grid.div_cos_lat((u, v))
    modal_outputs['vorticity'] = dinosaur_grid.curl_cos_lat((u, v))
    return modal_outputs


@dataclasses.dataclass
class ToNodalWithVelocity(TransformABC):
  """Module that converts inputs to nodal replacing div/curl with velocity."""

  grid: coordinates.SphericalHarmonicGrid | coordinates.LonLatGrid
  u_key: str = 'u_component_of_wind'
  v_key: str = 'v_component_of_wind'

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    dinosaur_grid = self.grid.ylm_grid
    if 'divergence' not in inputs or 'vorticity' not in inputs:
      raise ValueError(f'required `u, v` not found in {inputs.keys()=}')
    divergence, vorticity = inputs.pop('divergence'), inputs.pop('vorticity')
    u, v = spherical_harmonic.vor_div_to_uv_nodal(
        dinosaur_grid, vorticity=vorticity, divergence=divergence
    )
    to_nodal_fn = lambda x: dinosaur_grid.to_nodal(x) if x is not None else None
    nodal_outputs = jax.tree.map(to_nodal_fn, inputs)
    nodal_outputs[self.u_key] = u
    nodal_outputs[self.v_key] = v
    return nodal_outputs


class Nondimensionalize(TransformABC):
  """Transform that nondimensionalizes inputs."""

  def __init__(
      self,
      sim_units: units.SimUnits,
      inputs_to_units_mapping: dict[str, str],
  ):
    self.inputs_to_units_mapping = inputs_to_units_mapping
    self.sim_units = sim_units

  def _nondim_numeric(self, x: typing.Numeric, k: str):
    if k not in self.inputs_to_units_mapping:
      raise ValueError(f'Key {k} not found in {self.inputs_to_units_mapping=}')
    quantity = typing.Quantity(self.inputs_to_units_mapping[k])
    return self.sim_units.nondimensionalize(quantity * x)

  def _nondim_timed_field(self, x: data_specs.TimedField[cx.Field], k: str):
    nondim_value = self._nondim_numeric(x.field.data, k)
    return dataclasses.replace(x, field=cx.wrap_like(nondim_value, x.field))

  def _nondim_timed_observations(
      self, x: data_specs.TimedObservations[cx.Field]
  ) -> data_specs.TimedObservations:
    nondim_values = {
        k: self._nondim_numeric(v.data, k) for k, v in x.fields.items()
    }
    nondim_fields = {
        k: cx.wrap_like(v, x.fields[k]) for k, v in nondim_values.items()
    }
    return dataclasses.replace(x, fields=nondim_fields)

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    # here we handle different DataObservation types separately for now.
    result = {}
    for k, v in inputs.items():
      if isinstance(v, data_specs.TimedField):
        result[k] = self._nondim_timed_field(v, k)
      elif isinstance(v, data_specs.TimedObservations):
        result[k] = self._nondim_timed_observations(v)
      elif isinstance(v, typing.Numeric):
        result[k] = self._nondim_numeric(v, k)
      else:
        raise ValueError(f'Unsupported type {type(v)} for key {k}.')
    return from_dict_fn(result)


class Redimensionalize(TransformABC):
  """Transform that redimensionalizes inputs."""

  def __init__(
      self,
      sim_units: units.SimUnits,
      inputs_to_units_mapping: dict[str, str],
  ):
    self.inputs_to_units_mapping = inputs_to_units_mapping
    self.sim_units = sim_units

  def _redim_numeric(self, x: typing.Numeric, k: str):
    if k not in self.inputs_to_units_mapping:
      raise ValueError(f'Key {k} not found in {self.inputs_to_units_mapping=}')
    unit = typing.Quantity(self.inputs_to_units_mapping[k])
    return self.sim_units.dimensionalize(x, unit, as_quantity=False)

  def _redim_timed_field(self, x: data_specs.TimedField[cx.Field], k: str):
    dim_value = self._redim_numeric(x.field.data, k)
    return dataclasses.replace(x, field=cx.wrap_like(dim_value, x.field))

  def _redim_timed_observations(
      self, x: data_specs.TimedObservations[cx.Field]
  ) -> data_specs.TimedObservations:
    redim_values = {
        k: self._redim_numeric(v.data, k) for k, v in x.fields.items()
    }
    redim_fields = {
        k: cx.wrap_like(v, x.fields[k]) for k, v in redim_values.items()
    }
    return dataclasses.replace(x, fields=redim_fields)

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    # here we handle different DataObservation types separately for now.
    result = {}
    for k, v in inputs.items():
      if isinstance(v, data_specs.TimedField):
        result[k] = self._redim_timed_field(v, k)
      elif isinstance(v, data_specs.TimedObservations):
        result[k] = self._redim_timed_observations(v)
      elif isinstance(v, typing.Numeric):
        result[k] = self._redim_numeric(v, k)
      else:
        raise ValueError(f'Unsupported type {type(v)} for key {k}.')
    return from_dict_fn(result)


class ToModal(TransformABC):
  """Module that returns inputs that are converted from nodal to modal space."""

  def __init__(
      self,
      grid: coordinates.LonLatGrid | coordinates.SphericalHarmonicGrid,
  ):
    self.grid = grid

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    modal_outputs = {}
    for k, v in inputs.items():
      modal_outputs[k] = self.grid.ylm_grid.to_modal(v)
    return modal_outputs


class ToModalWithFilteredGradients(TransformABC):
  """Module that returns filtered grad and laplacian features of inputs fields.

  Gradients are filtered with an exponential filter of order 1 and provided
  attentuations. If no attentuations are provided, then this transform returns
  no gradient features. To avoid accidental accumulation of the cos(lat)
  factors, features must be keyed using typing.KeyWithCosLatFactor namedtuple.
  """

  def __init__(
      self,
      grid: coordinates.LonLatGrid | coordinates.SphericalHarmonicGrid,
      filter_attenuations: tuple[float, ...] = tuple(),
  ):
    self.grid = grid
    self.attenuations = filter_attenuations
    modal_filters = [
        spatial_filters.ExponentialModalFilter(grid, attenuation=a, order=1)
        for a in filter_attenuations
    ]
    self.modal_filters = modal_filters

  def __call__(
      self,
      inputs: dict[typing.KeyWithCosLatFactor, typing.Array],
  ) -> dict[typing.KeyWithCosLatFactor, typing.Array]:
    dinosaur_grid = self.grid.ylm_grid
    features = {}
    for k, value in inputs.items():
      name, cos_lat_order = k.name, k.factor_order
      for filter_module, att in zip(self.modal_filters, self.attenuations):
        d_value_dlon, d_value_dlat = dinosaur_grid.cos_lat_grad(value)
        laplacian_value = dinosaur_grid.laplacian(value)
        # since gradient values picked up cos_lat factor we increment the
        # corresponding key. This factor is adjusted at the caller level.
        dlon_key = typing.KeyWithCosLatFactor(
            name + f'_dlon_{att}', cos_lat_order + 1
        )
        dlat_key = typing.KeyWithCosLatFactor(
            name + f'_dlat_{att}', cos_lat_order + 1
        )
        del2_key = typing.KeyWithCosLatFactor(
            name + f'_del2_{att}', cos_lat_order
        )
        features[dlon_key] = filter_module.filter_modal(d_value_dlon)
        features[dlat_key] = filter_module.filter_modal(d_value_dlat)
        features[del2_key] = filter_module.filter_modal(laplacian_value)
    return features


def _scale_levels_for_matching_keys(
    inputs: typing.Pytree,
    scales: typing.Array,
    keys_to_scale: Sequence[str] = tuple(),
) -> typing.Pytree:
  """Transforms `inputs` by scaling levels for keys that are in `keys_to_scale.

  Args:
    inputs: pytree of values that will be selectively scaled along levels.
    scales: scaling weights that will be applied.
    keys_to_scale: keys for which scaling operation is applied.

  Returns:
    pytree of the same structure of `inputs` with keys matching `keys_to_scale`
    scaled by `scales` along the level axis.
  """
  if scales.ndim != 1:
    raise ValueError(
        'scales must be 1d array of weights per level, got '
        f'array with shape {scales.shape}'
    )
  scales = scales[:, np.newaxis, np.newaxis]  # broadcasting shape.
  inputs, from_dict_fn = pytree_utils.as_dict(inputs)
  scale_fn = lambda x: x * scales
  inputs = pytree_utils.map_over_matching_keys(inputs, scale_fn, keys_to_scale)
  return from_dict_fn(inputs)


class LevelScale(TransformABC):
  """Transform inputs by scaling different vertical levels."""

  def __init__(
      self,
      scales: Sequence[float],
      keys_to_scale: Sequence[str] = tuple(),
  ):
    self.scale_fn = functools.partial(
        _scale_levels_for_matching_keys,
        scales=np.asarray(scales),
        keys_to_scale=keys_to_scale,
    )

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    return self.scale_fn(inputs)


class InverseLevelScale(TransformABC):
  """Transform inputs by inverse scaling different vertical levels."""

  def __init__(
      self,
      scales: Sequence[float],
      keys_to_scale: Sequence[str] = tuple(),
  ):
    self.scale_fn = functools.partial(
        _scale_levels_for_matching_keys,
        scales=1 / np.asarray(scales),
        keys_to_scale=keys_to_scale,
    )

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    return self.scale_fn(inputs)


#
# Transforms that are return pytrees with injected features.
#


# TSI: Energy input to the top of the Earth's atmosphere
# https://www.ncei.noaa.gov/products/climate-data-records/total-solar-irradiance
TOTAL_SOLAR_IRRADIANCE = 1361 * typing.units.W / typing.units.meter**2
# Seasonal variation in apparent solar irradiance due to Earth-Sun distance
SOLAR_IRRADIANCE_VARIATION = (
    47 * typing.units.W / typing.units.meter**2
)  # .5 * 6.9% * TSI


class RadiationFeatures(TransformABC):
  """Feature module that computes incident radiation flux."""

  def __init__(
      self,
      coords: coordinates.DinosaurCoordinates,
      sim_units: units.SimUnits,
      transform: Transform = Identity(),
  ):
    self.coords = coords
    self.transform = transform
    self.orbital_rate = jax.tree.map(
        sim_units.nondimensionalize,
        radiation.OrbitalTime(
            2 * jnp.pi / typing.units.year, 2 * jnp.pi / typing.units.day
        ),
    )
    self.mean_irradiance = sim_units.nondimensionalize(TOTAL_SOLAR_IRRADIANCE)
    self.variation = sim_units.nondimensionalize(SOLAR_IRRADIANCE_VARIATION)
    datetime = radiation.datetime64_to_datetime(sim_units.reference_datetime)
    self.reference_orbital_time = radiation.datetime_to_orbital_time(datetime)

  @property
  def lon(self) -> typing.Array:
    return self.coords.dinosaur_grid.nodal_mesh[0]

  @property
  def lat(self) -> typing.Array:
    return np.arcsin(self.coords.dinosaur_grid.nodal_mesh[1])

  def time_to_orbital_time(self, time: typing.Numeric) -> radiation.OrbitalTime:
    """Returns the OribtalTime corresponding to the specified nondim time."""
    orbital_time = self.reference_orbital_time + self.orbital_rate * time
    # Reduce the magnitude of the result to avoid loss of precision errors
    # downstream. Avoid jnp.fmod, which is not very precise on float32.
    orbital_time -= orbital_time // (2 * jnp.pi) * (2 * jnp.pi)
    return orbital_time

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    features = {}
    orbital_time = self.time_to_orbital_time(inputs['sim_time'])
    features['radiation'] = radiation.get_normalized_radiation_flux(
        orbital_time=orbital_time,
        longitude=self.lon,
        latitude=self.lat,
        mean_irradiance=self.mean_irradiance,
        variation=self.variation,
    )
    features = jax.tree.map(lambda x: jnp.expand_dims(x, 0), features)
    return self.transform(features)


@dataclasses.dataclass
class LatitudeFeatures(TransformABC):
  """Feature module that creates cos and sin of latitude features."""

  grid: coordinates.SphericalHarmonicGrid | coordinates.LonLatGrid
  transform: Transform = Identity()

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    del inputs  # unused.
    _, sin_lat = self.grid.ylm_grid.nodal_mesh
    sin_features = sin_lat[np.newaxis, ...]
    cos_features = jnp.cos(jnp.arcsin(sin_features))
    features = {
        'cos_latitude': cos_features,
        'sin_latitude': sin_features.astype(cos_features.dtype),
    }
    return self.transform(features)


@dataclasses.dataclass
class OrographyFeatures(TransformABC):
  """Feature module that computes orographic features."""

  orography_module: orographies.ModalOrography | nnx.Module
  transform: Transform = Identity()

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    del inputs  # unused.
    return self.transform({
        'orography': self.orography_module.nodal_orography[np.newaxis, ...],
    })


@dataclasses.dataclass
class OrographyWithGradsFeatures(TransformABC):
  """Feature module that computes orographic features and their gradients."""

  orography_module: orographies.ModalOrography | nnx.Module
  compute_gradients_transform: ToModalWithFilteredGradients | None = None
  include_raw_orography: bool = True
  transform: Transform = Identity()

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    del inputs  # unused.
    modal_features = {
        'orography': self.orography_module.modal_orography[np.newaxis, ...],
    }
    modal_features = {  # jit should eliminate to_modal if it is not used.
        typing.KeyWithCosLatFactor(k, 0): v for k, v in modal_features.items()
    }
    modal_gradient_features = self.compute_gradients_transform(modal_features)
    sh_grid = self.orography_module.grid.ylm_grid
    sec_lat = 1 / sh_grid.cos_lat
    sec2_lat = sh_grid.sec2_lat
    sec_lat_scales = {0: 1, 1: sec_lat, 2: sec2_lat}
    features = {}
    if self.include_raw_orography:
      all_modal_features = modal_gradient_features | modal_features
    else:
      all_modal_features = modal_gradient_features
    for k, v in all_modal_features.items():
      sec_lat_scale = sec_lat_scales[k.factor_order]
      features[k.name] = sh_grid.to_nodal(v) * sec_lat_scale
    return self.transform(features)


@dataclasses.dataclass
class PressureFeatures(TransformABC):
  """Feature module that computes pressure."""

  coords: coordinates.DinosaurCoordinates
  transform: Transform = Identity()

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    del input_shapes  # unused.
    horizontal = self.coords.horizontal
    assert isinstance(horizontal, coordinates.SphericalHarmonicGrid)
    nodal_coords = coordinates.DinosaurCoordinates(
        horizontal=horizontal.to_lon_lat_grid(),
        vertical=self.coords.vertical,
    )
    return {'pressure': ShapeFloatStruct(nodal_coords.shape)}

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    to_nodal_fn = self.coords.dinosaur_grid.to_nodal
    sigma = self.coords.fields['sigma'].data
    surface_pressure = jnp.exp(to_nodal_fn(inputs['log_surface_pressure']))
    pressure = surface_pressure * sigma[:, jnp.newaxis, jnp.newaxis]
    return self.transform({'pressure': pressure})


@dataclasses.dataclass
class RandomnessFeatures(nnx.Module):
  """Feature module that returns values from a random process."""

  random_process: random_processes.RandomProcessModule
  grid: cx.Coordinate
  feature_name: str = 'randomness'
  transform: Transform = Identity()

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    del input_shapes  # unused.
    leading_shape = self.random_process.event_shape
    return {
        self.feature_name: ShapeFloatStruct(leading_shape + self.grid.shape)
    }

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    del inputs  # unused.
    # TODO(dkochkov) consider whether we need to expand dim for surface field.
    return self.transform({
        self.feature_name: self.random_process.state_values(self.grid).data,
    })


@dataclasses.dataclass
class DynamicInputFeatures(TransformABC):
  """Feature module for computes dynamic input features."""

  # TODO(dkochkov) Generalize this to work beyond surface features.

  keys: Sequence[str]
  dynamic_input_module: dynamic_io.DynamicInputSlice
  transform: Transform = Identity()

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    del input_shapes  # unused.
    data_shapes = self.dynamic_input_module.output_shapes()
    return {
        k: ShapeFloatStruct((1,) + v.shape) if v.ndim == 2 else v
        for k, v in data_shapes.items()
        if k in self.keys
    }

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    sim_time = inputs['sim_time']
    data_features = self.dynamic_input_module(sim_time)
    # TODO(dkochkov) Used to expand dim here.
    features = {k: data_features[k].data for k in self.keys}
    features = {
        k: jnp.expand_dims(v, 0) if v.ndim == 2 else v
        for k, v in features.items()
    }
    return self.transform(features)


# TODO(dkochkov) Generalize this to work on coords with different grids.
class SpatialSurfaceFeatures(TransformABC):
  """Features for atmospheric columns that are static in space and time."""

  def __init__(
      self,
      feature_sizes: dict[str, int],
      *,
      grid: cx.Coordinate,
      transform: Transform = Identity(),
      param_type: TransformParams | nnx.Param = TransformParams,
      initializer: nnx.Initializer = nnx.initializers.truncated_normal(),
      rngs: nnx.Rngs,
  ):
    self.transform = transform
    self.features = {
        k: param_type(initializer(rngs.params(), shape=((v,) + grid.shape)))
        for k, v in feature_sizes.items()
    }

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    del inputs  # unused.
    return self.transform({k: v.value for k, v in self.features.items()})

  def update_features_from_data(
      self,
      dataset: xarray.Dataset,
      sim_units: units.SimUnits,
      spatial_filter=None,
  ):
    """Updates `self.features` with filtered data from dataset."""
    # TODO(dkochkov) use units attr on dataset with default to `meter` here.
    if spatial_filter is None:
      spatial_filter = lambda x: x

    lon_name, lat_name = xarray_utils.get_longitude_latitude_names(dataset)
    # TODO(dkochkov) Add consistency verification for lon/lat.
    # lon, lat = (dataset[lon_name], dataset[lat_name])
    # xarray_utils.verify_grid_consistency(lon, lat, self.coords)
    for key, feature_value in self.features.items():
      if key in dataset:
        data = dataset[key].transpose(lon_name, lat_name)
        data_units = units.parse_units(data.attrs['units'])
        data = sim_units.nondimensionalize(data.values * data_units)
        data = spatial_filter(data)
        if np.ndim(data) != 3:
          data = data[np.newaxis, ...]
        self.features[key] = type(feature_value)(data)


class VelocityAndPrognosticsWithModalGradients(TransformABC):
  """Features module that returns prognostics + u,v and optionally gradients."""

  def __init__(
      self,
      coords: coordinates.DinosaurCoordinates,
      surface_field_names: tuple[str, ...] = tuple(),
      volume_field_names: tuple[str, ...] = tuple(),
      compute_gradients_transform: ToModalWithFilteredGradients | None = None,
      transform: Transform = Identity(),
      inputs_are_modal: bool = True,
      u_key: str = 'u_component_of_wind',
      v_key: str = 'v_component_of_wind',
  ):
    if compute_gradients_transform is None:
      compute_gradients_transform = Empty()
    self.coords = coords
    self.surface_field_names = surface_field_names
    self.volume_field_names = volume_field_names
    self.fields_to_include = surface_field_names + volume_field_names
    self.compute_gradients_transform = compute_gradients_transform
    self.transform = transform
    self.u_key = u_key
    self.v_key = v_key
    if inputs_are_modal:
      self.pre_process = lambda x: x
    else:
      self.pre_process = ToModal(self.coords.horizontal)

  def _extract_features(
      self,
      inputs: typing.Pytree,
      prefix: str = '',
  ) -> typing.Pytree:
    """Returns a nodal velocity and prognostic features."""
    # Note: all intermediate features have an explicit cos-lat factors in key.
    # These factors are removed in the `__call__` method before returning.

    dinosaur_grid = self.coords.dinosaur_grid
    # compute `u, v` if div/curl is available and `u, v` not in prognosics.
    if set(['vorticity', 'divergence']).issubset(inputs.keys()) and (
        not set([self.u_key, self.v_key]).intersection(inputs.keys())
    ):
      cos_lat_u, cos_lat_v = spherical_harmonic.get_cos_lat_vector(
          inputs['vorticity'], inputs['divergence'], dinosaur_grid
      )
      modal_features = {}
      if self.u_key in self.fields_to_include:
        modal_features[typing.KeyWithCosLatFactor(prefix + self.u_key, 1)] = (
            cos_lat_u
        )
      if self.v_key in self.fields_to_include:
        modal_features[typing.KeyWithCosLatFactor(prefix + self.v_key, 1)] = (
            cos_lat_v
        )
    elif self.u_key in inputs and self.v_key in inputs:
      modal_features = {
          typing.KeyWithCosLatFactor(prefix + self.u_key, 0): inputs[
              self.u_key
          ],
          typing.KeyWithCosLatFactor(prefix + self.v_key, 0): inputs[
              self.v_key
          ],
      }
    else:
      modal_features = {}
    prognostics_keys = list(inputs.keys())
    if 'tracers' in prognostics_keys:
      prognostics_keys.remove('tracers')
    if 'sim_time' in prognostics_keys:
      prognostics_keys.remove('sim_time')
    for k in set(self.fields_to_include) - set([self.u_key, self.v_key]):
      if k in prognostics_keys:
        modal_features[typing.KeyWithCosLatFactor(prefix + k, 0)] = inputs[k]
      elif k in inputs['tracers']:
        value = inputs['tracers'][k]
        modal_features[typing.KeyWithCosLatFactor(prefix + k, 0)] = value
      else:
        raise ValueError(f'Prognostic {k} not found in inputs.')

    # Computing gradient features and adjusting cos_lat factors.
    # TODO(dkochkov) reenable sharding constraints.
    dinosaur_grid = self.coords.dinosaur_grid
    # modal_features = self.coords.with_dycore_sharding(modal_features)
    diff_operator_features = self.compute_gradients_transform(modal_features)
    sec_lat = 1 / dinosaur_grid.cos_lat
    sec2_lat = dinosaur_grid.sec2_lat
    sec_lat_scales = {0: 1, 1: sec_lat, 2: sec2_lat}
    # Computing all features in nodal space.
    features = {}
    for k, v in (diff_operator_features | modal_features).items():
      sec_lat_scale = sec_lat_scales[k.factor_order]
      features[k.name] = dinosaur_grid.to_nodal(v) * sec_lat_scale
    # TODO(dkochkov) reenable sharding constraints.
    # features = self.coords.with_dycore_sharding(features)
    return features

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    del input_shapes  # unused.
    dinosaur_coords = self.coords.dinosaur_coords  # pytype: disable=attribute-error
    modal_surface_shape = ShapeFloatStruct(dinosaur_coords.surface_modal_shape)
    modal_volume_shape = ShapeFloatStruct(dinosaur_coords.modal_shape)
    out_shapes = {}
    for k in self.surface_field_names:
      out_shapes[typing.KeyWithCosLatFactor(k, 0)] = modal_surface_shape
    for k in set(self.volume_field_names):
      out_shapes[typing.KeyWithCosLatFactor(k, 0)] = modal_volume_shape
    # note: `cos_lat` factors might be off for `u`, `v`, but that doesn't affect
    # the keys/shape of the output.
    out_shapes |= self.compute_gradients_transform.output_shapes(out_shapes)
    out_shapes = {k.name: v for k, v in out_shapes.items()}
    out_shapes = coordinates.modal_shape_to_nodal_shape(
        out_shapes, self.coords.horizontal
    )
    return self.transform.output_shapes(out_shapes)

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    inputs = self.pre_process(inputs)
    nodal_features = self._extract_features(inputs)
    return self.transform(nodal_features)


class PrognosticFeatures(TransformABC):
  """Features module that returns prognostics variables as is."""

  def __init__(
      self,
      coords: coordinates.DinosaurCoordinates,
      surface_field_names: tuple[str, ...] = tuple(),
      volume_field_names: tuple[str, ...] = tuple(),
      transform: Transform = Identity(),
  ):
    self.coords = coords
    self.surface_field_names = surface_field_names
    self.volume_field_names = volume_field_names
    self.fields_to_include = surface_field_names + volume_field_names
    self.transform = transform

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    output_features = {}
    for k in self.fields_to_include:
      output_features[k] = inputs[k]
    return self.transform(output_features)

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    del input_shapes  # unused.
    surface_shape = ShapeFloatStruct((1,) + self.coords.horizontal.shape)
    volume_shape = ShapeFloatStruct(self.coords.shape)
    out_shapes = {}
    for k in self.surface_field_names:
      out_shapes[k] = surface_shape
    for k in self.volume_field_names:
      out_shapes[k] = volume_shape
    return self.transform.output_shapes(out_shapes)


class EmbeddedFeatures(TransformABC):
  """Features module that returns feature generated from an embedding module."""

  def __init__(
      self,
      fields_to_embed: Sequence[str],
      *,
      embedding_module: nnx.Module,
      transform: Transform = Identity(),
  ):
    self.fields_to_embed = fields_to_embed
    self.embedding_module = embedding_module
    self.transform = transform

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    output_features = self.embedding_module(inputs)
    return self.transform(output_features)

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    return self.transform.output_shapes(
        self.embedding_module.output_shapes(input_shapes)
    )


class SurfaceEmbeddingFeatures(TransformABC):
  """Features module that returns embeddings for each surface point."""

  def __init__(
      self,
      coords: coordinates.DinosaurCoordinates,
      embedding_sizes: dict[str, int],
      embedding_factory: nnx.Module,
      transform: Transform = Identity(),
  ):
    embedding_shapes = {
        k: typing.ShapeFloatStruct((v,) + coords.horizontal.shape)
        for k, v in embedding_sizes.items()
    }
    self.embedding = embedding_factory(output_shapes=embedding_shapes)
    self.transform = transform

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    output_features = self.embedding(inputs)
    return self.transform(output_features)

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    return self.transform.output_shapes(self.embedding.output_shapes)


class VolumeEmbeddingFeatures(TransformABC):
  """Features module that returns embeddings for each point on the coords."""

  def __init__(
      self,
      coords: cx.Coordinate,
      embedding_names: tuple[str, ...],
      embedding_factory: nnx.Module,
      transform: Transform = Identity(),
  ):
    embedding_shapes = {
        k: typing.ShapeFloatStruct(coords.shape) for k in embedding_names
    }
    self.embedding = embedding_factory(output_shapes=embedding_shapes)
    self.transform = transform

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    output_features = self.embedding(inputs)
    return self.transform(output_features)

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    return self.transform.output_shapes(self.embedding.output_shapes)


@dataclasses.dataclass
class CombinedFeatures(TransformABC):
  """Feature module that combines multiple feature modules together.

  Feature modules that will be combined are specified as dictionary values where
  keys indicate optional feature prefix. This helps with: (1) disambiguating
  multiple differently configured features; (2) accessing feature modules of a
  configured model. By default, prefix is only added if `always_add_prefix` is
  set to True or if there's a conflict in feature names.
  """

  feature_modules: dict[str, Transform]
  transform: Transform = Identity()
  always_add_prefix: bool = False

  def __call__(self, inputs: typing.Pytree) -> dict[str, typing.Array]:
    all_features = {}
    for name_prefix, feature_module in self.feature_modules.items():
      features = feature_module(inputs)
      for k, v in features.items():
        if k not in all_features and not self.always_add_prefix:
          feature_key = k
        else:
          feature_key = '_'.join([name_prefix, k])
          if feature_key in all_features:
            raise ValueError(f'Encountered duplicate {feature_key=}')
        all_features[feature_key] = v
    return self.transform(all_features)

  def output_shapes(
      self, input_shapes: typing.Pytree | None = None
  ) -> typing.Pytree:
    out_shapes = {}
    for m in self.feature_modules.values():
      out_shapes |= m.output_shapes(input_shapes)
    return self.transform.output_shapes(out_shapes)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import textwrap

from absl.testing import absltest
import jax
import jax.numpy as jnp
from neuralgcm.experimental.coordax import named_axes
import numpy as np


def assert_named_array_equal(
    actual: named_axes.NamedArray,
    expected: named_axes.NamedArray,
) -> None:
  """Asserts that a NamedArray has the expected data and dims."""
  np.testing.assert_array_equal(actual.data, expected.data)
  assert actual.dims == expected.dims, (expected.dims, actual.dims)


class NamedAxesTest(absltest.TestCase):

  def test_named_array(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, ('x', None))
    self.assertEqual(array.dims, ('x', None))
    np.testing.assert_array_equal(array.data, data)
    self.assertEqual(array.ndim, 2)
    self.assertEqual(array.shape, (2, 5))
    self.assertEqual(array.positional_shape, (5,))
    self.assertEqual(array.named_shape, {'x': 2})
    self.assertEqual(
        repr(array),
        textwrap.dedent("""\
            NamedArray(
                data=Array([[0, 1, 2, 3, 4],
                            [5, 6, 7, 8, 9]], dtype=int32),
                dims=('x', None),
            )"""),
    )

  def test_constructor_error(self):
    with self.assertRaisesRegex(
        ValueError, re.escape(r'data.ndim=2 != len(dims)=1')
    ):
      named_axes.NamedArray(np.zeros((2, 5)), ('x',))
    with self.assertRaisesRegex(
        ValueError, re.escape(r'dimension names may not be repeated')
    ):
      named_axes.NamedArray(np.zeros((2, 5)), ('x', 'x'))

  def test_constructor_no_dims(self):
    data = np.arange(10).reshape((2, 5))
    expected = named_axes.NamedArray(data, (None, None))
    actual = named_axes.NamedArray(data)
    assert_named_array_equal(actual, expected)

  def test_tree_map_same_dims(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, ('x', 'y'))
    actual = jax.tree.map(lambda x: x, array)
    assert_named_array_equal(actual, array)

  def test_tree_map_cannot_trim(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, ('x', 'y'))
    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            'cannot trim named dimensions when unflattening to a NamedArray:'
            " ('x',)."
        ),
    ):
      jax.tree.map(lambda x: x[0, :], array)

  def test_tree_map_wrong_dim_size(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, ('x', 'y'))
    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            'named shape mismatch when unflattening to a NamedArray: '
            "{'x': 2, 'y': 3} != {'x': 2, 'y': 5}."
        ),
    ):
      jax.tree.map(lambda x: x[:, :3], array)

  def test_tree_map_new_dim(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, ('x', 'y'))
    expected = named_axes.NamedArray(data[np.newaxis, ...], (None, 'x', 'y'))
    actual = jax.tree.map(lambda x: x[np.newaxis, ...], array)
    assert_named_array_equal(actual, expected)

  def test_tree_map_trim_dim(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, (None, 'y'))
    expected = named_axes.NamedArray(data[0, ...], ('y',))
    actual = jax.tree.map(lambda x: x[0, ...], array)
    assert_named_array_equal(actual, expected)

  def test_jit(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, ('x', 'y'))
    actual = jax.jit(lambda x: x)(array)
    assert_named_array_equal(actual, array)

  def test_vmap(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, (None, 'y'))

    def identity_with_checks(x):
      self.assertEqual(x.dims, ('y',))
      return x

    actual = jax.vmap(identity_with_checks)(array)
    assert_named_array_equal(actual, array)

    array = named_axes.NamedArray(data, ('x', 'y'))
    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            'If you are using vmap or scan, the first dimension must be'
            ' unnamed.'
        ),
    ):
      jax.vmap(lambda x: x)(array)

  def test_scan(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, (None, 'y'))
    _, actual = jax.lax.scan(lambda _, x: (None, x), init=None, xs=array)
    assert_named_array_equal(actual, array)

  def test_tag_valid(self):
    data = np.arange(10).reshape((2, 5))

    array = named_axes.NamedArray(data, (None, 'y'))
    expected = named_axes.NamedArray(data, ('x', 'y'))
    actual = array.tag('x')
    assert_named_array_equal(actual, expected)

    array = named_axes.NamedArray(data, (None, None))
    expected = named_axes.NamedArray(data, ('x', 'y'))
    actual = array.tag('x', 'y')
    assert_named_array_equal(actual, expected)

  def test_tag_errors(self):
    data = np.arange(10).reshape((2, 5))

    array = named_axes.NamedArray(data, (None, 'y'))
    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            'there must be exactly as many dimensions given to `tag` as there'
            ' are positional axes in the array, but got () for '
            '1 positional axis.'
        ),
    ):
      array.tag()

    array = named_axes.NamedArray(data, (None, None))
    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            'there must be exactly as many dimensions given to `tag` as there'
            " are positional axes in the array, but got ('x',) for "
            '2 positional axes.'
        ),
    ):
      array.tag('x')

    with self.assertRaisesRegex(
        TypeError,
        re.escape('dimension names must be strings: (None, None)'),
    ):
      array.tag(None, None)

  def test_untag_valid(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, ('x', 'y'))

    expected = named_axes.NamedArray(data, (None, 'y'))
    actual = array.untag('x')
    assert_named_array_equal(actual, expected)

    expected = named_axes.NamedArray(data, ('x', None))
    actual = array.untag('y')
    assert_named_array_equal(actual, expected)

    expected = named_axes.NamedArray(data, (None, None))
    actual = array.untag('x', 'y')
    assert_named_array_equal(actual, expected)

  def test_untag_invalid(self):
    data = np.arange(10).reshape((2, 5))
    partially_named_array = named_axes.NamedArray(data, (None, 'y'))
    fully_named_array = named_axes.NamedArray(data, ('x', 'y'))

    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            '`untag` cannot be used to introduce positional axes for a'
            ' NamedArray that already has positional axes. Please assign names'
            ' to the existing positional axes first using `tag`.'
        ),
    ):
      partially_named_array.untag('y')

    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            "cannot untag ('invalid',) because they are not a subset of the"
            " current named dimensions ('x', 'y')"
        ),
    ):
      fully_named_array.untag('invalid')

    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            "cannot untag ('y', 'x') because they do not appear in the order of"
            " the current named dimensions ('x', 'y')"
        ),
    ):
      fully_named_array.untag('y', 'x')

  def test_order_as(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, ('x', 'y'))

    actual = array.order_as('x', 'y')
    assert_named_array_equal(actual, array)

    actual = array.order_as('x', ...)
    assert_named_array_equal(actual, array)

    actual = array.order_as(..., 'y')
    assert_named_array_equal(actual, array)

    actual = array.order_as(...)
    assert_named_array_equal(actual, array)

    expected = named_axes.NamedArray(data.T, ('y', 'x'))
    actual = array.order_as('y', 'x')
    assert_named_array_equal(actual, expected)

    expected = named_axes.NamedArray(data.T, ('y', 'x'))
    actual = array.order_as('y', ...)
    assert_named_array_equal(actual, expected)

    expected = named_axes.NamedArray(data.T, ('y', 'x'))
    actual = array.order_as(..., 'x')
    assert_named_array_equal(actual, expected)

  def test_order_as_unnamed_dims(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, ('x', None))
    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            'cannot reorder the dimensions of an array with unnamed '
            "dimensions: ('x', None)"
        ),
    ):
      array.order_as('x', ...)

  def test_order_as_repeated_ellipsis(self):
    data = np.arange(10).reshape((2, 5))
    array = named_axes.NamedArray(data, ('x', 'y'))
    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            "dimension names contain multiple ellipses (...): (Ellipsis, 'x',"
            ' Ellipsis)'
        ),
    ):
      array.order_as(..., 'x', ...)

  def test_order_as_within_vmap(self):
    data = np.arange(10).reshape((1, 2, 5))
    array = named_axes.NamedArray(data, (None, 'x', 'y'))
    expected = named_axes.NamedArray(data.mT, (None, 'y', 'x'))
    actual = jax.vmap(lambda x: x.order_as('y', 'x'))(array)
    assert_named_array_equal(actual, expected)

  def test_nmap_identity(self):
    data = np.arange(2 * 3 * 4).reshape((2, 3, 4))

    def identity_assert_ndim(ndim):
      def f(x):
        self.assertIsInstance(x, jnp.ndarray)
        self.assertEqual(x.ndim, ndim)
        return x

      return f

    array = named_axes.NamedArray(data, (None, None, None))
    actual = named_axes.nmap(identity_assert_ndim(ndim=3))(array)
    assert_named_array_equal(actual, array)

    array = named_axes.NamedArray(data, ('x', 'y', 'z'))
    actual = named_axes.nmap(identity_assert_ndim(ndim=0))(array)
    assert_named_array_equal(actual, array)

    array = named_axes.NamedArray(data, ('x', 'y', None))
    expected = array.tag('z').order_as('z', ...).untag('z')
    actual = named_axes.nmap(identity_assert_ndim(ndim=1))(array)
    assert_named_array_equal(actual, expected)

  def test_nmap_scalar_only(self):
    expected = named_axes.NamedArray(3, ())
    actual = named_axes.nmap(jnp.add)(1, 2)
    assert_named_array_equal(actual, expected)

  def test_nmap_namedarray_and_scalar(self):
    data = np.arange(2 * 3 * 4).reshape((2, 3, 4))
    array = named_axes.NamedArray(data, ('x', 'y', 'z'))
    expected = named_axes.NamedArray(data + 1, ('x', 'y', 'z'))

    actual = named_axes.nmap(jnp.add)(array, 1)
    assert_named_array_equal(actual, expected)

    actual = named_axes.nmap(jnp.add)(1, array)
    assert_named_array_equal(actual, expected)

  def test_nmap_two_named_arrays(self):
    data = np.arange(2 * 3 * 4).reshape((2, 3, 4))

    array = named_axes.NamedArray(data, ('x', 'y', 'z'))
    expected = named_axes.NamedArray(data * 2, ('x', 'y', 'z'))
    actual = named_axes.nmap(jnp.add)(array, array)
    assert_named_array_equal(actual, expected)

    actual = named_axes.nmap(jnp.add)(array, array.order_as('z', 'y', 'x'))
    assert_named_array_equal(actual, expected)

    array1 = named_axes.NamedArray(data, ('x', 'y', 'z'))
    array2 = named_axes.NamedArray(100 * data[0, :, 0], ('y',))
    expected = named_axes.NamedArray(
        data + 100 * data[:1, :, :1], ('x', 'y', 'z')
    )
    actual = named_axes.nmap(jnp.add)(array1, array2)
    assert_named_array_equal(actual, expected)

    expected = expected.order_as('y', ...).untag('y')
    actual = named_axes.nmap(jnp.add)(array1.untag('y'), array2.untag('y'))
    assert_named_array_equal(actual, expected)

    array1 = named_axes.NamedArray(data[:, 0, 0], dims=('x',))
    array2 = named_axes.NamedArray(100 * data[0, :, 0], dims=('y',))
    expected = named_axes.NamedArray(
        data[:, :1, 0] + 100 * data[:1, :, 0], ('x', 'y')
    )
    actual = named_axes.nmap(jnp.add)(array1, array2)
    assert_named_array_equal(actual, expected)

    actual = named_axes.nmap(jnp.add)(array2, array1)
    assert_named_array_equal(actual, expected.order_as('y', 'x'))

  def test_nmap_axis_name(self):
    data = np.arange(2 * 3).reshape((2, 3))
    array = named_axes.NamedArray(data, ('x', 'y'))
    expected = named_axes.NamedArray(
        data - data.sum(axis=1, keepdims=True), ('x', 'y')
    )
    actual = named_axes.nmap(lambda x: x - x.sum(axis='y'))(array)
    assert_named_array_equal(actual, expected)

  def test_nmap_inconsistent_named_shape(self):

    def accepts_anything(*args, **kwargs):
      return 1

    array1 = named_axes.NamedArray(np.zeros((2, 3)), ('x', 'y'))
    array2 = named_axes.NamedArray(np.zeros((4,)), 'y')

    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            'Inconsistent sizes in a call to nmap(<NAME>) '
            "for dimensions ['y']:"
            "\n  args[0].named_shape == {'x': 2, 'y': 3}"
            "\n  args[1].named_shape == {'y': 4}"
        ).replace('NAME', '.+'),
    ):
      named_axes.nmap(accepts_anything)(array1, array2)

    array1 = named_axes.NamedArray(np.zeros((2, 3)), ('x', 'y'))
    array2 = named_axes.NamedArray(np.zeros((4, 5)), ('y', 'x'))

    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            'Inconsistent sizes in a call to nmap(<NAME>) '
            "for dimensions ['x', 'y']:"
            "\n  kwargs['bar'][0].named_shape == {'y': 4, 'x': 5}"
            "\n  kwargs['foo'].named_shape == {'x': 2, 'y': 3}"
        ).replace('NAME', '.+'),
    ):
      named_axes.nmap(accepts_anything)(foo=array1, bar=[array2])

  def test_nmap_out_axes_reorder(self):
    data = np.arange(2 * 3 * 4).reshape((2, 3, 4))
    array = named_axes.NamedArray(data, ('x', 'y', 'z'))

    expected = array.order_as('y', 'x', 'z')
    out_axes = {'y': 0, 'x': 1, 'z': 2}
    actual = named_axes.nmap(lambda x: x, out_axes=out_axes)(array)
    assert_named_array_equal(actual, expected)

    expected = array.order_as('y', 'x', 'z')
    out_axes = {'y': -3, 'x': -2, 'z': -1}
    actual = named_axes.nmap(lambda x: x, out_axes=out_axes)(array)
    assert_named_array_equal(actual, expected)

    expected = array.order_as('z', 'x', 'y')
    out_axes = {'z': 0, 'x': 1, 'y': 2}
    actual = named_axes.nmap(lambda x: x, out_axes=out_axes)(array)
    assert_named_array_equal(actual, expected)

  def test_nmap_out_axes_new_dim(self):
    data = np.arange(2 * 3 * 4).reshape((2, 3, 4))
    array = named_axes.NamedArray(data, ('x', 'y', 'z'))

    expected = named_axes.NamedArray(
        data[jnp.newaxis, ...], (None, 'x', 'y', 'z')
    )
    out_axes = {'x': 1, 'y': 2, 'z': 3}
    actual = named_axes.nmap(lambda x: x[jnp.newaxis], out_axes=out_axes)(array)
    assert_named_array_equal(actual, expected)

    expected = named_axes.NamedArray(
        data[:, jnp.newaxis, ...], ('x', None, 'y', 'z')
    )
    out_axes = {'x': 0, 'y': 2, 'z': 3}
    actual = named_axes.nmap(lambda x: x[jnp.newaxis], out_axes=out_axes)(array)
    assert_named_array_equal(actual, expected)

    expected = named_axes.NamedArray(
        data[..., jnp.newaxis], ('x', 'y', 'z', None)
    )
    out_axes = {'x': 0, 'y': 1, 'z': 2}
    actual = named_axes.nmap(lambda x: x[jnp.newaxis], out_axes=out_axes)(array)
    assert_named_array_equal(actual, expected)

    expected = named_axes.NamedArray(
        data[:, jnp.newaxis, ...], ('x', None, 'y', 'z')
    )
    out_axes = {'x': -4, 'y': -2, 'z': -1}
    actual = named_axes.nmap(lambda x: x[jnp.newaxis], out_axes=out_axes)(array)
    assert_named_array_equal(actual, expected)

    expected = named_axes.NamedArray(
        data.mT[..., jnp.newaxis], ('x', 'z', 'y', None)
    )
    out_axes = {'x': -4, 'z': -3, 'y': -2}
    actual = named_axes.nmap(lambda x: x[jnp.newaxis], out_axes=out_axes)(array)
    assert_named_array_equal(actual, expected)

  def test_nmap_out_binary(self):
    data1 = np.arange(2 * 3).reshape((2, 3))
    data2 = 10 * np.arange(3 * 2).reshape((3, 2))
    array1 = named_axes.NamedArray(data1, ('x', 'y'))
    array2 = named_axes.NamedArray(data2, ('y', 'x'))

    expected1 = array1
    expected2 = array2.order_as('x', 'y')
    actual1, actual2 = named_axes.nmap(lambda x, y: (x, y))(array1, array2)
    assert_named_array_equal(actual1, expected1)
    assert_named_array_equal(actual2, expected2)

    expected1 = array1
    expected2 = array2.order_as('x', 'y')
    actual1, actual2 = named_axes.nmap(
        lambda x, y: (x, y), out_axes={'x': 0, 'y': 1}
    )(array1, array2)
    assert_named_array_equal(actual1, expected1)
    assert_named_array_equal(actual2, expected2)

    expected1 = array1.order_as('y', 'x')
    expected2 = array2
    actual1, actual2 = named_axes.nmap(
        lambda x, y: (x, y), out_axes={'x': 1, 'y': 0}
    )(array1, array2)
    assert_named_array_equal(actual1, expected1)
    assert_named_array_equal(actual2, expected2)

  def test_nmap_invalid_out_axes(self):
    data = np.arange(2 * 3).reshape((2, 3))
    array = named_axes.NamedArray(data, ('x', 'y'))

    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            "out_axes keys ['x'] must must match the named "
            "dimensions ['x', 'y']"
        ),
    ):
      named_axes.nmap(lambda x: x, out_axes={'x': 0})(array)

    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            'out_axes must be either all positive or all negative, but got '
            "{'x': 0, 'y': -1}"
        ),
    ):
      named_axes.nmap(lambda x: x, out_axes={'x': 0, 'y': -1})(array)

    with self.assertRaisesRegex(
        ValueError,
        re.escape(
            'out_axes must all have unique values, but got '
            "{'x': 0, 'y': 0}"
        ),
    ):
      named_axes.nmap(lambda x: x, out_axes={'x': 0, 'y': 0})(array)

  def test_vectorized_methods(self):
    data = np.arange(2 * 3 * 4).reshape((2, 3, 4))
    dims = (None, 'y', 'z')
    array = named_axes.NamedArray(data, dims)

    expected = named_axes.NamedArray(-data, dims)
    actual = -array
    assert_named_array_equal(actual, expected)

    expected = named_axes.NamedArray(data - 1, dims)
    actual = array - 1
    assert_named_array_equal(actual, expected)

    expected = named_axes.NamedArray(1 - data, dims)
    actual = 1 - array
    assert_named_array_equal(actual, expected)

    expected = named_axes.NamedArray(data + 1j * data, dims)
    actual = (array - 1j * array).conj()
    assert_named_array_equal(actual, expected)

  def test_scalar_conversion(self):
    array = named_axes.NamedArray(1, dims=())
    expected = 1
    actual = int(array)
    self.assertIsInstance(actual, int)
    self.assertEqual(expected, actual)


if __name__ == '__main__':
  jax.config.update('jax_traceback_filtering', 'off')
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ML modules for equation-based models."""

from typing import Any, Callable, Optional, Sequence, Union
from dinosaur import coordinate_systems
from dinosaur import held_suarez
from dinosaur import primitive_equations
from dinosaur import pytree_utils
from dinosaur import scales
from dinosaur import shallow_water
from dinosaur import sigma_coordinates
from dinosaur import time_integration
from dinosaur import typing
from dinosaur import xarray_utils
import gin
import haiku as hk
import jax
import jax.numpy as jnp
from neuralgcm import features
from neuralgcm import mappings
from neuralgcm import orographies
from neuralgcm import parameterizations

units = scales.units
SCALE = scales.DEFAULT_SCALE
QuantityOrStr = Union[str, scales.Quantity]
EquationModule = Callable[..., time_integration.ImplicitExplicitODE]
TransformModule = typing.TransformModule
FeaturesModule = features.FeaturesModule
OrographyModule = orographies.OrographyModule
MappingModule = mappings.MappingModule
StepFilterModule = Callable[..., typing.PyTreeStepFilterFn]

REF_TEMP_KEY = xarray_utils.REF_TEMP_KEY
REF_POTENTIAL_KEY = xarray_utils.REF_POTENTIAL_KEY
OROGRAPHY = xarray_utils.OROGRAPHY


@gin.register
class ShallowWaterEquations(shallow_water.ShallowWaterEquations):
  """Equation module for shallow water system."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: shallow_water.ShallowWaterSpecs,
      aux_features: typing.AuxFeatures,
      orography_module: OrographyModule = orographies.ClippedOrography,
      name: Optional[str] = None,
  ):
    reference_potential = aux_features.get(REF_POTENTIAL_KEY, None)
    if reference_potential is None:
      raise ValueError(f'must supply {REF_POTENTIAL_KEY} in `aux_features`.')
    modal_orography_init_fn = orography_module(
        coords, dt, physics_specs, aux_features)
    modal_orography = modal_orography_init_fn()  # pytype: disable=not-callable  # jax-ndarray
    super().__init__(
        coords=coords,
        physics_specs=physics_specs,
        orography=modal_orography,
        reference_potential=reference_potential,
    )


@gin.register
class PrimitiveEquations(primitive_equations.PrimitiveEquations):
  """Equation module for primitive equations."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: primitive_equations.PrimitiveEquationsSpecs,
      aux_features: typing.AuxFeatures,
      orography_module: OrographyModule = orographies.ClippedOrography,
      vertical_advection: Callable[..., jax.Array] = (
          sigma_coordinates.centered_vertical_advection
      ),
      include_vertical_advection: bool = True,
      name: Optional[str] = None,
  ):
    ref_temperatures = aux_features.get(REF_TEMP_KEY, None)
    if ref_temperatures is None:
      raise ValueError(f'must supply {REF_TEMP_KEY} in `aux_features`.')
    modal_orography_init_fn = orography_module(
        coords, dt, physics_specs, aux_features)
    modal_orography = modal_orography_init_fn()  # pytype: disable=not-callable  # jax-ndarray
    super().__init__(
        coords=coords,
        physics_specs=physics_specs,
        reference_temperature=ref_temperatures,
        orography=modal_orography,
        vertical_advection=vertical_advection,
        include_vertical_advection=include_vertical_advection,
    )


@gin.register
class PrimitiveEquationsWithTime(
    primitive_equations.PrimitiveEquationsWithTime
):
  """Equation module for primitive equations."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: primitive_equations.PrimitiveEquationsSpecs,
      aux_features: typing.AuxFeatures,
      orography_module: OrographyModule = orographies.ClippedOrography,
      vertical_advection: Callable[..., jax.Array] = (
          sigma_coordinates.centered_vertical_advection
      ),
      include_vertical_advection: bool = True,
      name: Optional[str] = None,
  ):
    ref_temperatures = aux_features.get(REF_TEMP_KEY, None)
    if ref_temperatures is None:
      raise ValueError(f'must supply {REF_TEMP_KEY} in `aux_features`.')
    modal_orography_init_fn = orography_module(
        coords, dt, physics_specs, aux_features)
    modal_orography = modal_orography_init_fn()  # pytype: disable=not-callable  # jax-ndarray
    super().__init__(
        coords=coords,
        physics_specs=physics_specs,
        reference_temperature=ref_temperatures,
        orography=modal_orography,
        vertical_advection=vertical_advection,
        include_vertical_advection=include_vertical_advection,
    )


@gin.register
class MoistPrimitiveEquations(
    primitive_equations.MoistPrimitiveEquations
):
  """Equation module for moist primitive equations."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: primitive_equations.PrimitiveEquationsSpecs,
      aux_features: typing.AuxFeatures,
      orography_module: OrographyModule = orographies.ClippedOrography,
      vertical_advection: Callable[..., jax.Array] = (
          sigma_coordinates.centered_vertical_advection
      ),
      include_vertical_advection: bool = True,
      name: Optional[str] = None,
  ):
    ref_temperatures = aux_features.get(REF_TEMP_KEY, None)
    if ref_temperatures is None:
      raise ValueError(f'must supply {REF_TEMP_KEY} in `aux_features`.')
    modal_orography_init_fn = orography_module(
        coords, dt, physics_specs, aux_features)
    modal_orography = modal_orography_init_fn()  # pytype: disable=not-callable  # jax-ndarray
    super().__init__(
        coords=coords,
        physics_specs=physics_specs,
        reference_temperature=ref_temperatures,
        orography=modal_orography,
        vertical_advection=vertical_advection,
        include_vertical_advection=include_vertical_advection,
    )


@gin.register
class MoistPrimitiveEquationsWithCloudMoisture(
    primitive_equations.MoistPrimitiveEquationsWithCloudMoisture
):
  """Equation module for moist primitive equations with clouds."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: primitive_equations.PrimitiveEquationsSpecs,
      aux_features: typing.AuxFeatures,
      orography_module: OrographyModule = orographies.ClippedOrography,
      vertical_advection: Callable[..., jax.Array] = (
          sigma_coordinates.centered_vertical_advection
      ),
      include_vertical_advection: bool = True,
      name: Optional[str] = None,
  ):
    ref_temperatures = aux_features.get(REF_TEMP_KEY, None)
    if ref_temperatures is None:
      raise ValueError(f'must supply {REF_TEMP_KEY} in `aux_features`.')
    modal_orography_init_fn = orography_module(
        coords, dt, physics_specs, aux_features)
    modal_orography = modal_orography_init_fn()  # pytype: disable=not-callable  # jax-ndarray
    super().__init__(
        coords=coords,
        physics_specs=physics_specs,
        reference_temperature=ref_temperatures,
        orography=modal_orography,
        vertical_advection=vertical_advection,
        include_vertical_advection=include_vertical_advection,
    )


@gin.register
class MoistPrimitiveEquationsWithCloudMoisutre(
    MoistPrimitiveEquationsWithCloudMoisture
):
  """Temporary alias with mis-spelled name."""


@gin.register
class HeldSuarezEquations(held_suarez.HeldSuarezForcing):
  """Equation module for Held-Suarez forcing equations."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: primitive_equations.PrimitiveEquationsSpecs,
      aux_features: typing.AuxFeatures,
      name: Optional[str] = None,
  ):
    ref_temperatures = aux_features.get(REF_TEMP_KEY, None)
    if ref_temperatures is None:
      raise ValueError(f'must supply {REF_TEMP_KEY} in `aux_features`.')
    super().__init__(
        coords=coords,
        physics_specs=physics_specs,
        reference_temperature=ref_temperatures)


# TODO(dkochkov) Test if vertical diffusion works well with euler integrator.


@gin.register
class VerticalDiffusion(time_integration.ExplicitODE):
  """Equation module that adds explicit diffusion along vertical direction."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      timescale: QuantityOrStr = gin.REQUIRED,
  ):
    self.coords = coords
    timescale = dt / physics_specs.nondimensionalize(scales.Quantity(timescale))
    timescales = coords.vertical.boundaries * timescale
    self.level_weighted_timescales = timescales[:, jnp.newaxis, jnp.newaxis]

  def explicit_terms(self, state: typing.PyTreeState) -> typing.PyTreeState:
    def vertical_diffusion_fn(x: typing.Array) -> typing.Array:
      # TODO(dkochkov) Consider using sigma_coordinates.centered_difference.
      x_grad = x[1:, ...] - x[:-1, ...]
      # padding with zero values for vertical fluxes.
      pad_width = ((1, 1), (0, 0), (0, 0))
      x_grad = jnp.pad(x_grad, pad_width)
      fluxes = self.level_weighted_timescales * x_grad
      # TODO(dkochkov) Consider using sigma_coordinates.centered_difference.
      return fluxes[1:, ...] - fluxes[:-1, ...]

    nodal_state = self.coords.horizontal.to_nodal(state)
    nodal_tendency = pytree_utils.tree_map_where(
        condition_fn=lambda x: jnp.asarray(x).shape == self.coords.nodal_shape,
        f=vertical_diffusion_fn,
        g=jnp.zeros_like,
        x=nodal_state)
    modal_tendency = self.coords.horizontal.to_modal(nodal_tendency)
    return self.coords.horizontal.clip_wavenumbers(modal_tendency)


@gin.register
class NoDynamics(time_integration.ImplicitExplicitODE):
  """The constant ODE, ∂u/∂t = 0."""

  def __init__(self, *args, **kwargs):
    del args, kwargs

  def explicit_terms(self, x: typing.PyTreeState) -> typing.PyTreeState:
    return 0 * x

  def implicit_terms(self, x: typing.PyTreeState) -> typing.PyTreeState:
    return 0 * x

  def implicit_inverse(
      self, x: typing.PyTreeState, time_step: float
  ) -> typing.PyTreeState:
    return x


@gin.register
def composed_equations_module(
    coords: coordinate_systems.CoordinateSystem,
    dt: float,
    physics_specs: Any,
    aux_features: typing.AuxFeatures,
    equation_modules: Sequence[EquationModule],
) -> time_integration.ImplicitExplicitODE:
  """Returns an equation module that represents a composition of equations."""
  equations = tuple(eq(coords, dt, physics_specs, aux_features)
                    for eq in equation_modules)
  return time_integration.compose_equations(equations)


@gin.register
class DirectNeuralEquations(hk.Module, time_integration.ExplicitODE):
  """Computes explicit tendencies for the input state.

  This equation module predicts tendencies directly in the nodal representation
  and returns values transformed back to the modal space. The nodal tendencies
  are computed by the `nodal_mapping_module` from preprocessed nodal features
  computed by `modal_to_nodal_features_module` followed by the
  `tendency_transform_module`.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      modal_to_nodal_features_module: FeaturesModule,
      nodal_mapping_module: mappings.MappingModule,
      tendency_transform_module: TransformModule,
      prediction_mask: Optional[typing.Pytree] = None,
      filter_module: Optional[StepFilterModule] = None,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.parameterization_fn = parameterizations.DirectNeuralParameterization(
        coords=coords,
        dt=dt,
        physics_specs=physics_specs,
        aux_features=aux_features,
        modal_to_nodal_features_module=modal_to_nodal_features_module,
        nodal_mapping_module=nodal_mapping_module,
        tendency_transform_module=tendency_transform_module,
        prediction_mask=prediction_mask,
        filter_module=filter_module,
        name=name,
    )

  def explicit_terms(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    modal_tendencies = self.parameterization_fn(inputs, forcing=None)
    modal_tendencies = pytree_utils.none_to_zeros(modal_tendencies, inputs)
    return modal_tendencies


@gin.register
class DivCurlNeuralEquations(hk.Module, time_integration.ExplicitODE):
  """Computes explicit tendencies using div and curl operators for `u, v` terms.

  This equation module predicts tendencies of the inputs with velocity-based
  parameterization of the `divergence` and `vorticity` components. Specifically,
  we replace predictions of `divergence` and `vorticity` by nodal predictions
  of `u`, and `v`, which are then differentiated using modal representation.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      modal_to_nodal_features_module: FeaturesModule,
      nodal_mapping_module: mappings.MappingModule,
      tendency_transform_module: TransformModule,
      prediction_mask: Optional[typing.Pytree] = None,
      filter_module: Optional[StepFilterModule] = None,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.parameterization_fn = parameterizations.DivCurlNeuralParameterization(
        coords=coords,
        dt=dt,
        physics_specs=physics_specs,
        aux_features=aux_features,
        modal_to_nodal_features_module=modal_to_nodal_features_module,
        nodal_mapping_module=nodal_mapping_module,
        tendency_transform_module=tendency_transform_module,
        prediction_mask=prediction_mask,
        filter_module=filter_module,
        name=name,
    )

  def explicit_terms(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    modal_tendencies = self.parameterization_fn(inputs, forcing=None)
    modal_tendencies = pytree_utils.none_to_zeros(modal_tendencies, inputs)
    return modal_tendencies



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for layers.py."""

from typing import Optional, Sequence, Tuple

from absl.testing import absltest
from absl.testing import parameterized
import gin
import haiku as hk
import jax
from neuralgcm import layers  # pylint: disable=unused-import
import numpy as np


class LayersTest(parameterized.TestCase):
  """Tests layers primitives."""

  @parameterized.parameters(
      dict(
          output_size=3,
          num_hidden_units=6,
          num_hidden_layers=7,
          with_bias=True,),
      dict(
          output_size=5,
          num_hidden_units=13,
          num_hidden_layers=4,
          with_bias=False,),
  )
  def test_mlp_uniform_shape(
      self,
      output_size: int,
      num_hidden_units: int,
      num_hidden_layers: int,
      with_bias: bool,
  ):
    """Tests that EpdTower outputs and params have expected shapes."""
    gin.enter_interactive_mode()

    @gin.configurable
    def layer_forward(inputs, layer):  # defines a forward function
      return layer(output_size)(inputs)

    config = '\n'.join([
        'layer_forward.layer = @MlpUniform',
        f'MlpUniform.num_hidden_units = {num_hidden_units}',
        f'MlpUniform.num_hidden_layers = {num_hidden_layers}',
        f'MlpUniform.with_bias = {with_bias}',
        'ConvLonLat.name = "mlp_uniform"',
    ])

    gin.clear_config()
    gin.parse_config(config)

    n_features = 8
    input_size = (n_features,)
    inputs = np.random.uniform(size=input_size)
    model = hk.without_apply_rng(hk.transform(layer_forward))
    params = model.init(jax.random.PRNGKey(42), inputs)
    out = model.apply(params, inputs)
    with self.subTest('output_shape'):
      self.assertEqual(out.shape, (output_size,))
    with self.subTest('total_params_count'):
      expected = count_mlp_uniform_params(
          input_size=n_features,
          num_hidden_units=num_hidden_units,
          num_hidden_layers=num_hidden_layers,
          output_size=output_size,
          with_bias=with_bias,)
      actual = sum([np.prod(x.shape)
                    for x in jax.tree_util.tree_leaves(params)])
      self.assertEqual(actual, expected)
    with self.subTest('parameter_keys'):
      n_layers = num_hidden_layers + 1
      self.assertEqual(
          tuple(params.keys()),
          tuple([f'mlp_uniform/~/linear_{i}' for i in range(n_layers)]))

  @parameterized.parameters(
      dict(
          output_size=2,
          kernel_shape=(3, 3),
          with_bias=False,),
      dict(
          output_size=4,
          kernel_shape=(5, 5),
          with_bias=True,),
  )
  def test_conv_lon_lat_shape(
      self,
      output_size: int,
      kernel_shape: int,
      with_bias: bool,
  ):
    """Tests that ConvLonLat outputs and params have expected shapes."""
    gin.enter_interactive_mode()

    @gin.configurable
    def layer_forward(inputs, layer):  # defines a forward function
      return layer(output_size)(inputs)

    config = '\n'.join([
        'layer_forward.layer = @ConvLonLat',
        f'ConvLonLat.kernel_shape = {kernel_shape}',
        f'ConvLonLat.with_bias = {with_bias}',
        'ConvLonLat.name = "conv_lon_lat"',
    ])

    gin.clear_config()
    gin.parse_config(config)

    n_features = 8
    input_size = (n_features, 10, 20)
    inputs = np.random.uniform(size=input_size)
    model = hk.without_apply_rng(hk.transform(layer_forward))
    params = model.init(jax.random.PRNGKey(42), inputs)
    out = model.apply(params, inputs)
    with self.subTest('output_shape'):
      self.assertEqual(out.shape, (output_size, 10, 20))
    with self.subTest('total_params_count'):
      expected = count_conv_lon_lat_params(
          input_size=n_features,
          kernel_shape=kernel_shape,
          output_size=output_size,
          with_bias=with_bias)
      actual = sum([np.prod(x.shape)
                    for x in jax.tree_util.tree_leaves(params)])
      self.assertEqual(actual, expected)

  @parameterized.parameters(
      dict(
          output_channels=2,
          kernel_shape=3,
          ),
      dict(
          output_channels=4,
          kernel_shape=5,
          ),
    )
  def test_conv1d_shape(
      self,
      output_channels: int,
      kernel_shape: int,
  ):
    """Tests that ConvLonLat outputs and params have expected shapes."""
    gin.enter_interactive_mode()

    @gin.configurable
    def layer_forward(inputs, layer):  # defines a forward function
      return layer()(inputs)

    config = '\n'.join([
        'layer_forward.layer = @ConvLevel',
        f'ConvLevel.kernel_shape = {kernel_shape}',
        f'ConvLevel.output_channels = {output_channels}',
        'ConvLevel.name = "conv_sigma"',
    ])

    gin.clear_config()
    gin.parse_config(config)

    in_channels = 8
    input_shape = (in_channels, 24)
    key = jax.random.PRNGKey(42)
    inputs = jax.random.uniform(key, shape=input_shape)
    model = hk.without_apply_rng(hk.transform(layer_forward))
    params = model.init(key, inputs)
    out = model.apply(params, inputs)
    with self.subTest('output_shape'):
      self.assertEqual(out.shape, (output_channels, 24))
    with self.subTest('total_params_count'):
      expected = count_conv_lon_lat_params(
          input_size=in_channels,
          kernel_shape=kernel_shape,
          output_size=output_channels,
          with_bias=True)
      actual = sum([np.prod(x.shape)
                    for x in jax.tree_util.tree_leaves(params)])
      self.assertEqual(actual, expected)

  @parameterized.parameters(
      dict(
          output_channels=2,
          channels=[6],
          kernel_shapes=[3, 3],
          dilation_rates=[1, 1],
      ),
      dict(
          output_channels=4,
          channels=[7],
          kernel_shapes=[5, 3],
          dilation_rates=[3, 1],
      ),
  )
  def test_cnn_level_shape(
      self,
      output_channels: int,
      channels: Sequence[int],
      kernel_shapes: Sequence[int],
      dilation_rates: Sequence[int],
  ):
    """Tests that VerticalConvNet outputs and params have expected shapes."""
    gin.enter_interactive_mode()

    @gin.configurable
    def layer_forward(inputs, layer):  # defines a forward function
      return layer(output_channels)(inputs)

    config = '\n'.join([
        'layer_forward.layer = @VerticalConvNet',
        f'VerticalConvNet.channels = {channels}',
        f'VerticalConvNet.kernel_shapes = {kernel_shapes}',
        f'VerticalConvNet.dilation_rates = {dilation_rates}',
    ])

    gin.clear_config()
    gin.parse_config(config)

    in_channels = 8
    n_levels = 12
    input_shape = (in_channels, n_levels)
    key = jax.random.PRNGKey(42)
    inputs = jax.random.uniform(key, shape=input_shape)
    model = hk.without_apply_rng(hk.transform(layer_forward))
    params = model.init(key, inputs)
    out = model.apply(params, inputs)
    with self.subTest('output_shape'):
      actual_shape = out.shape
      expected_shape = (output_channels, n_levels)
      self.assertEqual(actual_shape, expected_shape)

  @parameterized.parameters(
      dict(
          inputs_shape=(3, 12),  # self-attention on d=3 inputs, returning 4.
          latents_shape=None,
          pos_encoding_shape=None,
          expected_output_shape=(4, 12),
          ),
      dict(
          inputs_shape=(3, 7),  # attending to latents on 12 levels.
          latents_shape=(2, 12),
          pos_encoding_shape=None,
          expected_output_shape=(5, 7),  # results on the same levels as inputs.
          ),
      dict(
          inputs_shape=(3, 8),  # attending to latents on 7 levels.
          latents_shape=(2, 7),
          pos_encoding_shape=(1, 8),
          expected_output_shape=(4, 8),  # results on the same levels as inputs.
          ),
      dict(
          inputs_shape=(3, 6),
          latents_shape=(2, 7),
          pos_encoding_shape=None,
          expected_output_shape=(8, 6),
          # can skip projection because output_size == latent_size.
          extra_config_str='LevelTransformer.skip_final_projection = True',
          ),
      dict(
          inputs_shape=(8, 6),
          latents_shape=(2, 7),
          pos_encoding_shape=None,
          expected_output_shape=(1, 6),
          # can skip input projection because input_shape[0] == latent_size.
          extra_config_str='LevelTransformer.input_projection_net = None',
          ),
    )
  def test_level_transformer_shape(
      self,
      inputs_shape: Tuple[int, int],
      latents_shape: Optional[Tuple[int, int]],
      pos_encoding_shape: Optional[Tuple[int, int]],
      expected_output_shape: Tuple[int, int],
      extra_config_str: str = '',
  ):
    """Tests that LevelTransformer outputs have expected shapes."""
    gin.enter_interactive_mode()
    output_channels, _ = expected_output_shape

    @gin.configurable
    def layer_forward(inputs, latents, pos_encoding, layer):
      return layer(output_channels)(inputs, latents, pos_encoding)

    config = '\n'.join([
        'layer_forward.layer = @LevelTransformer',
        'LevelTransformer.n_layers = 1',
        'LevelTransformer.latent_size = 8',
        'LevelTransformer.num_heads = 2',
        'LevelTransformer.key_size = 4',
    ])
    gin.clear_config()
    gin.parse_config(config)
    gin.parse_config(extra_config_str)

    rng_seq = hk.PRNGSequence(42)
    inputs = jax.random.uniform(next(rng_seq), shape=inputs_shape)
    if latents_shape is not None:
      latents = jax.random.uniform(next(rng_seq), shape=latents_shape)
    else:
      latents = None
    if pos_encoding_shape is not None:
      pos_encoding = jax.random.uniform(next(rng_seq), shape=pos_encoding_shape)
    else:
      pos_encoding = None

    model = hk.without_apply_rng(hk.transform(layer_forward))
    params = model.init(next(rng_seq), inputs, latents, pos_encoding)
    out = model.apply(params, inputs, latents, pos_encoding)
    with self.subTest('output_shape'):
      self.assertEqual(out.shape, expected_output_shape)


def count_mlp_uniform_params(
    input_size: int,
    num_hidden_units: int,
    num_hidden_layers: int,
    output_size: int,
    with_bias: bool):
  """Returns the number of parameters in MlpUniform based on settings."""
  n_input = input_size * num_hidden_units + num_hidden_units * with_bias
  n_hidden = num_hidden_units * num_hidden_units + num_hidden_units * with_bias
  n_output = num_hidden_units * output_size + output_size * with_bias
  return n_input + n_hidden * (num_hidden_layers - 1) + n_output


def count_conv_lon_lat_params(
    input_size: int,
    kernel_shape: int,
    output_size: int,
    with_bias: bool):
  """Returns the number of parameters in ConvLonLat based on settings."""
  n_w_params = np.prod(kernel_shape) * input_size * output_size
  n_b_params = output_size
  return n_w_params + n_b_params * with_bias

if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Shared utilities and classes for metrics and related modules."""
from __future__ import annotations
import dataclasses
from typing import Callable, Optional, Sequence

from dinosaur import coordinate_systems
from dinosaur import horizontal_interpolation
from dinosaur import pytree_utils
from dinosaur import spherical_harmonic
from dinosaur import typing
import frozendict
import gin
import jax
import jax.extend as jex
from jax.interpreters import ad
from jax.interpreters import batching
from jax.interpreters import mlir
import jax.numpy as jnp
import numpy as np


tree_map = jax.tree_util.tree_map
tree_leaves = jax.tree_util.tree_leaves
Pytree = typing.Pytree
TrajectoryRepresentations = typing.TrajectoryRepresentations

# Number of state variables in the model. t/z/u/v/specific_humidity.
N_VARS = 5


# Axis names.
TIME = 'time'
LEVEL = 'level'
LONGITUDINAL_WAVENUMBER = 'longitudinal_wavenumber'
TOTAL_WAVENUMBER = 'total_wavenumber'
LONGITUDINAL = 'longitudinal'
LATITUDINAL = 'latitudinal'

# SPATIAL_AXES is negatively indexed because it is used in a place where there
# are variable number of leading axis.
SPATIAL_AXES = (-2, -1)
TIME_AXIS = 0
LEVEL_AXIS = 1
ALL_AXES = (TIME_AXIS, LEVEL_AXIS) + SPATIAL_AXES


MODAL_AXIS_INDICES = frozendict.frozendict({
    TIME: TIME_AXIS,
    LEVEL: LEVEL_AXIS,
    LONGITUDINAL_WAVENUMBER: 2,
    TOTAL_WAVENUMBER: 3,
})


NODAL_AXIS_INDICES = frozendict.frozendict({
    TIME: TIME_AXIS,
    LEVEL: LEVEL_AXIS,
    LONGITUDINAL: 2,
    LATITUDINAL: 3,
})


class ShapeError(Exception):
  """Raised when an unexpected shape is encountered."""


@dataclasses.dataclass
class TrajectorySpec:
  """Specification of a saved model trajectory."""

  trajectory_length: int  # i.e., max "outer steps"
  max_trajectory_length: int  # Maximum length for any stage of an Experiment.
  steps_per_save: int  # Number of (1 hr) inner steps between each outer step.
  coords: coordinate_systems.CoordinateSystem  # i.e., model coords
  data_coords: coordinate_systems.CoordinateSystem  # i.e., data coords

  def __post_init__(self):
    if self.trajectory_length > self.max_trajectory_length:
      raise ValueError(
          f'{self.trajectory_length=} > {self.max_trajectory_length=}.'
      )


@dataclasses.dataclass
class TrajectoryShape:
  """Specifies shape of trajectory after LinearTransforms are applied."""

  n_times: int
  n_levels: int
  n_longitudinal_wavenumbers: int
  n_total_wavenumbers: int
  n_longitude_nodes: int
  n_latitude_nodes: int

  def assert_compliant(self, trajectory: typing.Pytree, is_nodal: bool) -> None:
    """Asserts `trajectory` is compliant with this `TrajectoryShape`.

    Args:
      trajectory: A trajectory, after LinearTransforms have been applied.
      is_nodal: Whether the trajectory is presumed nodal (vs. modal).

    Raises:
      ShapeError: If the shape is not compliant.
    """
    if is_nodal:
      expected_shape = (
          self.n_times,
          self.n_levels,
          self.n_longitude_nodes,
          self.n_latitude_nodes,
      )
    else:
      expected_shape = (
          self.n_times,
          self.n_levels,
          self.n_longitudinal_wavenumbers,
          self.n_total_wavenumbers,
      )

    is_compliant = tree_map(lambda x: np.shape(x) == expected_shape, trajectory)
    if not all(tree_leaves(is_compliant)):
      shapes = tree_map(np.shape, trajectory)
      raise ShapeError(
          f'Some `trajectory` shapes were non-compliant ({is_nodal=}). '
          f'{expected_shape=}. Found {shapes=}. '
          f'This TrajectoryShape is {self}.'
      )


def nodal_surface_mean(
    x: typing.Array, coords: coordinate_systems.CoordinateSystem
) -> typing.Array:
  """Integrates x over the surface of a sphere, normalized by surface area."""
  if x.shape[-2:] != coords.horizontal.nodal_shape[-2:]:
    raise ValueError(f'Input to nodal_surface_mean: {x.shape=}, while expected '
                     f'spatial shape is {coords.horizontal.nodal_shape=}.')
  surface_area = 4 * jnp.pi * coords.horizontal.radius**2
  # Changes shape (n_t, n_z, n_lon, n_lat) --> (n_t, n_z)
  return coords.horizontal.integrate(x) / surface_area


def modal_surface_mean(
    x: typing.Array, coords: coordinate_systems.CoordinateSystem
) -> typing.Array:
  """Integrates Σxₖφₖ² over a sphere, normalized by surface area."""
  if x.shape[-2:] != coords.horizontal.modal_shape[-2:]:
    raise ValueError(f'Input to modal_surface_mean: {x.shape=}, while expected '
                     f'modal shape is {coords.horizontal.modal_shape=}.')
  # This is equivalent to computing ||f||² / SurfaceArea, where
  #   f = Σₖsqrt(x)ₖφₖ
  surface_area = 4 * jnp.pi * coords.horizontal.radius**2

  # Changes shape (n_t, n_z, m, l) --> (n_t, n_z)
  return jnp.sum(x, axis=SPATIAL_AXES) / surface_area


def extract_time_slice(trajectory: Pytree, time_slice: slice) -> Pytree:
  return pytree_utils.slice_along_axis(trajectory, TIME_AXIS, time_slice)


def extract_time_step(trajectory: Pytree, time_step: int) -> Pytree:
  return extract_time_slice(trajectory, slice(time_step, time_step + 1))


def extract_vertical_slice(
    trajectory: Pytree,
    coords: coordinate_systems.CoordinateSystem,
    level: int,
) -> Pytree:
  i = coords.vertical.centers.tolist().index(level)
  index = slice(i, i + 1)
  trajectory = pytree_utils.slice_along_axis(trajectory, LEVEL_AXIS, index)
  return trajectory


def filter_sim_time(trajectory: Pytree) -> Pytree:
  if isinstance(trajectory, dict):
    trajectory = dict(trajectory)
    trajectory.pop('sim_time', None)
  return trajectory


def filter_sim_time_and_diagnostics(trajectory: Pytree) -> Pytree:
  if isinstance(trajectory, dict):
    trajectory = dict(trajectory)
    trajectory.pop('sim_time', None)
    trajectory.pop('diagnostics', None)
  return trajectory


def extract_variable(
    trajectory: TrajectoryRepresentations,
    trajectory_spec: TrajectorySpec,
    time_step: int | slice | None = None,
    level: int | None = None,
    getter: Callable[[Pytree], Pytree] = filter_sim_time,
    is_nodal: bool = True,
    is_encoded: bool = False,
) -> Pytree:
  """Extract a variable from a trajectory."""
  if is_encoded:
    coords = trajectory_spec.coords
  else:
    coords = trajectory_spec.data_coords
  trajectory = trajectory.get_representation(
      is_nodal=is_nodal, is_encoded=is_encoded
  )
  trajectory = getter(trajectory)
  if time_step is not None:
    if isinstance(time_step, slice):
      trajectory = extract_time_slice(trajectory, time_step)
    else:
      trajectory = extract_time_step(trajectory, time_step)
  if level is not None:
    trajectory = extract_vertical_slice(trajectory, coords, level)
  return trajectory


def replace_with_linear_trucation(
    trajectory_spec: TrajectorySpec,
) -> TrajectorySpec:
  """Replaces TrajectorySpec with a TL* version of it."""
  grid = trajectory_spec.data_coords.horizontal
  max_wavenumber = grid.longitude_wavenumbers - 1
  assert max_wavenumber + 2 == grid.total_wavenumbers
  gaussian_nodes = grid.longitude_nodes // 4
  assert gaussian_nodes == grid.latitude_nodes // 2

  # pytype: disable=attribute-error
  new_horizontal = spherical_harmonic.Grid.construct(
      max_wavenumber=2 * gaussian_nodes - 1,  # Larger in TL version
      gaussian_nodes=gaussian_nodes,  # Same in T and TL versions
      latitude_spacing=grid.latitude_spacing,
      radius=grid.radius,
  )
  # pytype: enable=attribute-error

  return dataclasses.replace(
      trajectory_spec,
      data_coords=dataclasses.replace(
          trajectory_spec.data_coords,
          horizontal=new_horizontal,
      ),
  )


def trajectory_4d_shape(
    trajectory_spec: TrajectorySpec,
    keep_levels: Optional[Sequence[float]] = None,
) -> TrajectoryShape:
  """Returns the shape of the trajectory leaf values in data representation."""
  if keep_levels is None:
    n_levels = trajectory_spec.data_coords.vertical.layers
  else:
    n_levels = sum(bool(i) for i in keep_levels)
    if n_levels > trajectory_spec.data_coords.vertical.layers:
      raise ValueError(
          f'{n_levels=} implied by `keep_levels` was greater than '
          f'{trajectory_spec.data_coords.vertical.layers=}'
      )
  grid = trajectory_spec.data_coords.horizontal
  n_m, n_l = grid.modal_shape
  return TrajectoryShape(
      n_times=trajectory_spec.trajectory_length,
      n_levels=n_levels,
      n_longitudinal_wavenumbers=n_m,
      n_total_wavenumbers=n_l,
      n_longitude_nodes=grid.longitude_nodes,
      n_latitude_nodes=grid.latitude_nodes,
  )


def pmean_all_axes(x: jax.Array) -> jax.Array:
  """Average over all vmapped axes."""
  return _pmean_all_axes_p.bind(x)


def _pmean_all_axes_impl(x):
  return x


def _pmean_all_axes_batch(args, batch_axes):
  (x,) = args
  (batch_axis,) = batch_axes
  y = jnp.broadcast_to(x.mean(axis=batch_axes, keepdims=True), x.shape)
  return _pmean_all_axes_p.bind(y), batch_axis


_pmean_all_axes_p = jex.core.Primitive('pmean_all_axes')
_pmean_all_axes_p.def_impl(_pmean_all_axes_impl)
_pmean_all_axes_p.def_abstract_eval(_pmean_all_axes_impl)
batching.primitive_batchers[_pmean_all_axes_p] = _pmean_all_axes_batch
ad.deflinear(_pmean_all_axes_p, lambda cotangent: [pmean_all_axes(cotangent)])
mlir.register_lowering(
    _pmean_all_axes_p,
    mlir.lower_fun(_pmean_all_axes_impl, multiple_results=False),
)


@dataclasses.dataclass
class AggregationTransform:
  """A transformation that aggregates spatial or temporal groups in inputs.

  These transformations are useful for (1) coarsening of error observations and
  (2) aggregation of error norms to compute L2^2 distance between two vectors.
  The former case does not strictly impose any restrictions on the coarsening
  transformation, although in most cases we would expect it to be a form of a
  linear, non-invertible transformation. The latter requires that the result of
  aggregation of non-negative values is non-negative.
  """

  trajectory_spec: TrajectorySpec
  out_trajectory_spec: TrajectorySpec
  is_nodal: bool
  is_encoded: bool

  def __call__(self, inputs: Pytree) -> Pytree:
    raise NotImplementedError


AggregationTransformConstructor = Callable[..., AggregationTransform]


@gin.register
class AggregateIdentity(AggregationTransform):

  def __init__(
      self,
      trajectory_spec: TrajectorySpec,
      is_nodal: bool,
      is_encoded: bool,
  ):
    super().__init__(trajectory_spec, trajectory_spec, is_nodal, is_encoded)

  def __call__(self, inputs: Pytree) -> Pytree:
    return inputs


@gin.register
class SumVariables(AggregationTransform):
  """Transform that adds sums all variables aka pytree leaves of inputs."""

  def __init__(
      self,
      trajectory_spec: TrajectorySpec,
      is_nodal: bool,
      is_encoded: bool,
  ):
    super().__init__(trajectory_spec, trajectory_spec, is_nodal, is_encoded)

  def __call__(self, inputs: Pytree) -> Pytree:
    return sum(jax.tree_util.tree_leaves(inputs))


@gin.register
class RegriddingAggregation(AggregationTransform):
  """Transform that aggregates horizontal cells via regridding.

  To perform aggregation over a few nearby lon/lat cells this transform performs
  regridding to a coarser `target_grid`. By default, the aggregated value
  contains a regridded (i.e. mean) value of the inputs. Setting `scale_by_area`
  to `True` multiplies outputs by an area which is close to area-weighted
  aggregation.
  """

  def __init__(
      self,
      trajectory_spec: TrajectorySpec,
      is_nodal: bool,
      is_encoded: bool,
      target_grid: coordinate_systems.CoordinateSystem,
      scale_by_area: bool = False,
  ):
    if not is_nodal:
      raise ValueError('AggregateHorizontal is only supported on nodal data')
    if is_encoded:
      source_coords = trajectory_spec.coords
      coords = dataclasses.replace(source_coords, horizontal=target_grid)  # pytype: disable=wrong-arg-types  # dataclasses-replace-types
      out_trajectory_spec = dataclasses.replace(trajectory_spec, coords=coords)
    else:
      source_coords = trajectory_spec.data_coords
      coords = dataclasses.replace(source_coords, horizontal=target_grid)  # pytype: disable=wrong-arg-types  # dataclasses-replace-types
      out_trajectory_spec = dataclasses.replace(
          trajectory_spec, data_coords=coords)
    super().__init__(trajectory_spec, out_trajectory_spec, is_nodal, is_encoded)
    self.regrid_fn = horizontal_interpolation.ConservativeRegridder(
        source_coords.horizontal, coords.horizontal)
    # conservative regridding computes weighted averages rather than aggregation
    # so we reweight the results by area.
    lower_lon_boundaries = horizontal_interpolation._periodic_lower_bounds(
        coords.horizontal.longitudes, 2 * np.pi)
    upper_lon_boundaries = horizontal_interpolation._periodic_upper_bounds(
        coords.horizontal.longitudes, 2 * np.pi)
    lat_boundaries = horizontal_interpolation._latitude_cell_bounds(
        coords.horizontal.latitudes)
    lon_weights = upper_lon_boundaries - lower_lon_boundaries
    lat_weights = jnp.sin(lat_boundaries[1:]) - jnp.sin(lat_boundaries[:-1])
    self.weights = lat_weights[np.newaxis, :] * lon_weights[:, np.newaxis]
    self.scale_by_area = scale_by_area

  def __call__(self, inputs: Pytree) -> Pytree:
    if self.scale_by_area:
      return tree_map(lambda x: self.regrid_fn(x) * self.weights, inputs)
    else:
      return tree_map(self.regrid_fn, inputs)


@gin.register
class TimeWindowSum(AggregationTransform):
  """Transform that sums temporal blocks of `time_window_size`."""

  def __init__(
      self,
      trajectory_spec: TrajectorySpec,
      is_nodal: bool,
      is_encoded: bool,
      time_window_size: int,
  ):
    trajectory_length = trajectory_spec.trajectory_length
    if trajectory_length % time_window_size != 0:
      raise ValueError(f'Cannot aggregate {trajectory_length=} '
                       f'into {time_window_size=} sections.')
    new_length = trajectory_spec.trajectory_length // time_window_size
    out_trajectory_spec = dataclasses.replace(
        trajectory_spec,
        trajectory_length=new_length,
        steps_per_save=trajectory_spec.steps_per_save * time_window_size)
    super().__init__(trajectory_spec, out_trajectory_spec, is_nodal, is_encoded)
    eye = np.eye(trajectory_length)
    # columns of the weight matrix have 1s in rows that are in the same window.
    # see http://screen/8CaZoBVNPtjpwNu for a hint.
    self.time_axis_weights = sum(
        [np.roll(eye, i, 0) for i in range(time_window_size)]
    )[:, ::time_window_size]

  def __call__(self, inputs: Pytree) -> Pytree:
    def _aggregate_time(x: jax.Array):
      return jnp.einsum(
          'tk,...thml->...khml', self.time_axis_weights, x, precision='float32')
    return tree_map(_aggregate_time, inputs)



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Types used by neuralgcm.experimental API."""
from __future__ import annotations

import abc
import dataclasses
import datetime
import functools
from typing import Any, Callable, Generic, TypeVar

import jax
import jax.numpy as jnp
from neuralgcm.experimental import scales
import numpy as np
import pandas as pd
import tree_math


units = scales.units
#
# Generic types.
#
Array = np.ndarray | jax.Array
Dtype = jax.typing.DTypeLike | Any
Numeric = float | int | Array
PRNGKeyArray = jax.Array
ShapeDtypeStruct = jax.ShapeDtypeStruct
ShapeFloatStruct = functools.partial(ShapeDtypeStruct, dtype=jnp.float32)
Timestep = np.timedelta64 | float
TimedeltaLike = str | np.timedelta64 | pd.Timestamp | datetime.timedelta
Quantity = scales.Quantity

#
# Generic API input/output types.
#
Pytree = Any
PyTreeState = TypeVar('PyTreeState')


#
# Simulation function signatures.
#
StepFn = Callable[[PyTreeState], PyTreeState]
PostProcessFn = Callable[..., Pytree]


class ObservationSpecs(abc.ABC):
  """Base class for observation data specifications.

  Observation specification objects are used to express queries to the model's
  `observe` method. All ObservationSpecs objects should have a corresponding
  ObservationData class that can be used to store the data, both of which must
  be a valid jax pytree.

  In many cases the difference between ObservationSpecs and ObservationData is
  that ObservationSpecs contain supporting values and a set of
  coordax.Coordinate objects while ObservationData is structured similarly but
  uses coordax.Field to store data. This can be reflected by using the same
  object that inherits from both ObservationSpecs and ObservationData and
  supports coordax.Field and coordax.Coordinate for relevant attributes.
  """


class ObservationData(abc.ABC):
  """Base class for observation data."""

  @property
  @abc.abstractmethod
  def get_specs(self) -> ObservationSpecs:
    """Returns ObservationSpecs that describe data specifications."""
    ...


Query = dict[str, ObservationSpecs]
Observation = dict[str, ObservationData]


@tree_math.struct
class ModelState(Generic[PyTreeState]):
  """Simulation state decomposed into prognostic, diagnostic and randomness.

  Attributes:
    prognostics: Prognostic variables describing the simulation state.
    diagnostics: Optional diagnostic values holding diagnostic information.
    randomness: Optional randomness state describing stochasticity of the model.
  """

  prognostics: PyTreeState
  diagnostics: Pytree = dataclasses.field(default_factory=dict)
  randomness: Pytree = dataclasses.field(default_factory=dict)


@jax.tree_util.register_pytree_node_class
@dataclasses.dataclass
class Randomness:
  """State describing the random process."""

  prng_key: jax.Array
  prng_step: int = 0
  core: Pytree = None

  def tree_flatten(self):
    """Flattens Randomness JAX pytree."""
    leaves = (self.prng_key, self.prng_step, self.core)
    aux_data = ()
    return leaves, aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, leaves):
    """Unflattens Randomness from aux_data and leaves."""
    return cls(*leaves, *aux_data)


@jax.tree_util.register_pytree_node_class
@dataclasses.dataclass
class Timedelta:
  """JAX compatible time duration, stored in days and seconds.

  Like datetime.timedelta, the Timedelta constructor and arithmetic operations
  normalize seconds to fall in the range [0, 24 * 60 * 60). Timedelta objects
  are pytrees, but normalization is skipped inside jax.tree operations because
  JAX uses pytrees with non-numeric types to implement JAX transformations.

  Using integer days and seconds is recommended to avoid loss of precision. With
  int32 days, Timedelta can exactly represent durations over 5 million years.

  The easiest way to create a Timedelta is to use `from_timedelta64`, which
  supports `np.timedelta64` objects and NumPy arrays with a timedelta64 dtype:

    >>> Timedelta.from_timedelta64(np.timedelta64(1, 's'))
    Timedelta(days=0, seconds=1)
  """

  days: Numeric = 0
  seconds: Numeric = 0

  # TODO(shoyer): can we rewrite this a custom JAX dtype, like jax.random.key?

  def __post_init__(self):
    days_delta, seconds = divmod(self.seconds, 24 * 60 * 60)
    self.days = self.days + days_delta
    self.seconds = seconds

  @classmethod
  def from_timedelta64(cls, values: np.timedelta64 | np.ndarray) -> Timedelta:
    seconds = values // np.timedelta64(1, 's')
    # no need to worry about overflow, because timedelta64 represents values
    # internally with int64 and normalization uses native array operations
    return Timedelta(0, seconds)

  def to_timedelta64(self) -> np.timedelta64 | np.ndarray:
    seconds = np.int64(self.days) * 24 * 60 * 60 + np.int64(self.seconds)
    return seconds * np.timedelta64(1, 's')

  def __add__(self, other):
    if not isinstance(other, Timedelta):
      return NotImplemented
    days = self.days + other.days
    seconds = self.seconds + other.seconds
    return Timedelta(days, seconds)

  def __neg__(self):
    return Timedelta(-self.days, -self.seconds)

  def __sub__(self, other):
    if not isinstance(other, Timedelta):
      return NotImplemented
    return self + (-other)

  def __mul__(self, other):
    if not isinstance(other, Numeric):
      return NotImplemented
    return Timedelta(self.days * other, self.seconds * other)

  __rmul__ = __mul__

  # TODO(shoyer): consider adding other methods supported by datetime.timedelta.

  def tree_flatten(self):
    leaves = (self.days, self.seconds)
    aux_data = None
    return leaves, aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, leaves):
    assert aux_data is None
    # JAX uses non-numeric values for pytree leaves inside transformations, so
    # we skip __post_init__ by constructing the object directly:
    # https://jax.readthedocs.io/en/latest/pytrees.html#custom-pytrees-and-initialization
    result = object.__new__(cls)
    result.days, result.seconds = leaves
    return result


_UNIX_EPOCH = np.datetime64('1970-01-01T00:00:00', 's')


@jax.tree_util.register_pytree_node_class
@dataclasses.dataclass
class Timestamp:
  """JAX compatible timestamp, stored as a delta from the Unix epoch.

  The easiest way to create a Timestamp is to use `from_datetime64`, which
  supports `np.datetime64` objects and NumPy arrays with a datetime64 dtype:

    >>> Timestamp.from_datetime64(np.datetime64('1970-01-02'))
    Timestamp(delta=Timedelta(days=1, seconds=0))
  """

  delta: Timedelta

  @classmethod
  def from_datetime64(cls, values: np.datetime64 | np.ndarray) -> Timestamp:
    return cls(Timedelta.from_timedelta64(values - _UNIX_EPOCH))

  def to_datetime64(self) -> np.timedelta64 | np.ndarray:
    return self.delta.to_timedelta64() + _UNIX_EPOCH

  def __add__(self, other):
    if not isinstance(other, Timedelta):
      return NotImplemented
    return Timestamp(self.delta + other)

  __radd__ = __add__

  def __sub__(self, other):
    if isinstance(other, Timestamp):
      return self.delta - other.delta
    elif isinstance(other, Timedelta):
      return Timestamp(self.delta - other)
    else:
      return NotImplemented

  def tree_flatten(self):
    leaves = (self.delta,)
    aux_data = None
    return leaves, aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, leaves):
    assert aux_data is None
    return cls(*leaves)


#
# API function signatures.
#
PostProcessFn = Callable[..., Any]


#
# Auxiliary types for intermediate computations.
#
@dataclasses.dataclass(eq=True, order=True, frozen=True)
class KeyWithCosLatFactor:
  """Class describing a key by `name` and an integer `factor_order`."""

  name: str
  factor_order: int



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper methods for constructing augmented trajectory functions."""

import dataclasses
import functools
from typing import Any, Callable, Sequence, Tuple
from dinosaur import coordinate_systems
from dinosaur import pytree_utils
from dinosaur import typing
import haiku as hk
import jax
import jax.numpy as jnp

Array = typing.Array
DynamicalSystem = Any  # to prevent circular dependency on model_builder
Pytree = typing.Pytree

tree_map = jax.tree_util.tree_map
tree_leaves = jax.tree_util.tree_leaves

# Linter confused by wrapped functions
# pylint: disable=g-bare-generic


def with_preprocessing(
    fn: Callable[..., Pytree],
    preprocess_fn: Callable,
) -> Callable[..., Pytree]:
  """Generates a function that computes `fn` on `preprocess_fn(x)`."""
  @functools.wraps(fn)
  def apply_fn(x, *args, **kwargs):
    return fn(preprocess_fn(x), *args, **kwargs)

  return apply_fn


def with_post_processing(
    fn: Callable[..., Pytree],
    post_process_fn: Callable,
) -> Callable[..., Pytree]:
  """Generates a function that applies `post_process_fn` to outputs of `fn`."""
  @functools.wraps(fn)
  def apply_fn(*args, **kwargs):
    return post_process_fn(*fn(*args, **kwargs))

  return apply_fn


def with_forcing(
    fn: Callable[..., Pytree],
    forcing_fn: typing.ForcingFn,
    forcing_data: typing.ForcingData,
) -> Callable[..., Pytree]:
  """Converts fn(x, forcing) to fn(x)."""
  # evaluates forcing=forcing_fn(forcing_data, x.sim_time)
  # when x does not have sim_time, forcing_fn will get sim_time=None
  @functools.wraps(fn)
  def wrapped(x, forcing_fn=forcing_fn):
    # handle dataclass or dict for state data
    if dataclasses.is_dataclass(x):
      if isinstance(x, typing.ModelState):
        sim_time = dataclasses.asdict(x.state).get('sim_time', None)
      else:
        sim_time = dataclasses.asdict(x).get('sim_time', None)
    else:
      sim_time = x.get('sim_time', None)
    # handle sim_time of ndim 0 or 1
    if sim_time is not None:
      sim_time = jax.numpy.asarray(sim_time)
      if sim_time.ndim:
        forcing_fn = jax.vmap(forcing_fn, in_axes=(None, 0))
    forcing = forcing_fn(forcing_data, sim_time)
    return fn(x, forcing=forcing)
  return wrapped


def with_split_input(
    fn: Callable[..., Pytree],
    split_index: int,
    time_axis: int = 0,
) -> Callable[..., Pytree]:
  """Decorates `fn` to be evaluated on first `split_index` time slices.

  The returned function is a generalization to pytrees of the function:
  `fn(x[:split_index], *args, **kwargs)`

  Args:
    fn: function to be transformed.
    split_index: number of input elements along the time axis to use.
    time_axis: axis corresponding to time dimension in `x` to decorated `fn`.

  Returns:
    decorated `fn` that is evaluated on only `split_index` first time slices of
    provided inputs.
  """
  @functools.wraps(fn)
  def apply_fn(x, *args, **kwargs):
    init, _ = pytree_utils.split_along_axis(x, split_index, axis=time_axis)
    return fn(init, *args, **kwargs)

  return apply_fn


def with_input_included(
    trajectory_fn: typing.TrajectoryFn,
    time_axis: int = 0,
    num_last_input_frames_to_trim: int = 0,
) -> typing.TrajectoryFn:
  """Returns a `trajectory_fn` that concatenates inputs `x` to trajectory."""
  if num_last_input_frames_to_trim > 0:
    num_last_input_frames_to_trim = -num_last_input_frames_to_trim
  else:
    num_last_input_frames_to_trim = None
  inputs_time_slice = slice(None, num_last_input_frames_to_trim)
  @functools.wraps(trajectory_fn)
  def _trajectory(x, *args, **kwargs):
    final, unroll = trajectory_fn(x, *args, **kwargs)
    x_concat = pytree_utils.slice_along_axis(x, time_axis, inputs_time_slice)
    return final, pytree_utils.concat_along_axis([x_concat, unroll], time_axis)

  return _trajectory


def trajectory_with_inputs_and_forcing(
    model: DynamicalSystem,
    num_init_frames: int,
    start_with_input: bool = False,
) -> typing.TrajectoryFn:
  """Returns trajectory_fn that comuptes model trajectory from target data.

  Wraps the default model.trajectory_fn to operate on data representation. It
  corresponds to slicing `num_init_frames` from the inputs, encoding and
  unrolling the trajectory.

  Args:
    model: model of a dynamical system used to obtain the trajectory.
    num_init_frames: number of time frames used from the physics trajectory to
      initialize the model state.
    start_with_input: whether the firest decoded step in the output trajectory
      should correspond to last input time or first future output.

  Returns:
    Trajectory function that operates on target data trajectory by encoding
    the `initial_frames` inputs and unrolls trajectory in a model space.
  """
  def _trajectory_fn(x, forcing_data, outer_steps, inner_steps=1):

    # configure the model.trajectory function with a decoder on the output.
    trajectory_fn = functools.partial(
        model.trajectory,
        outer_steps=outer_steps,
        inner_steps=inner_steps,
        forcing_data=forcing_data,
        start_with_input=start_with_input)
    # add preprocessing to encode input to model state.
    encode_fn = with_forcing(model.encode, model.forcing_fn, forcing_data)
    trajectory_fn = with_preprocessing(trajectory_fn, encode_fn)
    trajectory_fn = with_split_input(trajectory_fn, num_init_frames)
    return trajectory_fn(x)

  return _trajectory_fn


def trajectory_with_inputs_and_forcing_and_stop_gradients(
    model: DynamicalSystem,
    num_init_frames: int,
    start_with_input: bool = False,
    stop_gradient_outer_steps: Sequence[int] = (),
) -> typing.TrajectoryFn:
  """Returns trajectory_fn that comuptes model trajectory from target data.

  This extension of `trajectory_with_inputs_and_forcing` allows adding stop
  gradients to the trajectory at designated steps. For example, if
  `stop_gradient_outer_steps = [2]`, then gradients along the trajectory stop
  at t=2. This does not mean that gradients with respect to X[2] will be zero.
  It simply means that, for t > 2, gradients of X[t] with respect to X[2] will
  be zero.

  Wraps the default model.trajectory_fn to operate on data representation. It
  corresponds to slicing `num_init_frames` from the inputs, encoding and
  unrolling the trajectory.

  Args:
    model: model of a dynamical system used to obtain the trajectory.
    num_init_frames: number of time frames used from the physics trajectory to
      initialize the model state.
    start_with_input: whether the firest decoded step in the output trajectory
      should correspond to last input time or first future output.
    stop_gradient_outer_steps: Tuple (possibly empty) indicating outer steps at
      which to place stop gradients.

  Returns:
    Trajectory function that operates on target data trajectory by encoding
    the `initial_frames` inputs and unrolls trajectory in a model space.
    Decoding is not done by this function.
  """
  stop_gradient_outer_steps = list(sorted(stop_gradient_outer_steps))
  if num_init_frames != 1:
    raise ValueError(f'{num_init_frames=} is not supported yet.')

  if stop_gradient_outer_steps and min(stop_gradient_outer_steps) <= 0:
    raise ValueError(
        f'{stop_gradient_outer_steps=} contained non-positive values'
    )

  expand_dim0 = lambda tree: tree_map(lambda x_i: x_i[jnp.newaxis], tree)
  concat_dim0 = lambda trees: pytree_utils.concat_along_axis(trees, axis=0)
  slice_dim0 = lambda tree, idx: pytree_utils.slice_along_axis(
      tree, axis=0, idx=idx
  )

  def concat_trajectories_with_stop_grads(
      x, forcing_data, outer_steps, inner_steps=1
  ):
    if (
        stop_gradient_outer_steps
        and max(stop_gradient_outer_steps) > outer_steps
    ):
      raise ValueError(
          f'{stop_gradient_outer_steps=} contained values > {outer_steps=}'
      )
    outer_steps_seq = list(stop_gradient_outer_steps)
    if not outer_steps_seq or outer_steps_seq[-1] != outer_steps:
      outer_steps_seq.append(outer_steps)

    # The first leg needs to encode the input. So use
    # trajectory_with_inputs_and_forcing, which does the encoding.
    final_state, first_leg = trajectory_with_inputs_and_forcing(
        model,
        num_init_frames=num_init_frames,
        start_with_input=start_with_input,
    )(
        x,
        forcing_data=forcing_data,
        outer_steps=outer_steps_seq[0],
        inner_steps=inner_steps,
    )

    # At this point, sections contains times [0, ..., outer_steps_seq[0]]
    sections = [
        first_leg,
    ]

    # Subsequent legs do not need encoding, so use model.trajectory directly.
    trajectory_fn = functools.partial(
        model.trajectory,
        inner_steps=inner_steps,
        forcing_data=forcing_data,
        start_with_input=start_with_input,
    )
    for i in range(1, len(outer_steps_seq)):
      # outer_steps_seq[-1] may or may not be in stop_gradient_outer_steps.
      # The other steps will be by construction.
      assert set(outer_steps_seq[:-1]).issubset(stop_gradient_outer_steps)
      stop_grad_at_start = outer_steps_seq[i - 1] in stop_gradient_outer_steps

      initial_state = final_state

      # this_leg contains times [outer_steps_seq[0]+1, ..., outer_steps_seq[1]]
      final_state, this_leg = trajectory_fn(
          jax.lax.stop_gradient(initial_state)
          if stop_grad_at_start
          else initial_state,
          outer_steps=outer_steps_seq[i] - outer_steps_seq[i - 1],
      )

      if stop_grad_at_start and start_with_input:
        # Replace the initial point that had a stop gradient on it.
        this_leg = concat_dim0([
            expand_dim0(initial_state),
            slice_dim0(this_leg, idx=slice(1, None)),
        ])
      sections.append(this_leg)

    return final_state, concat_dim0(sections)

  return concat_trajectories_with_stop_grads


def decoded_trajectory_with_forcing(
    model: DynamicalSystem,
    start_with_input: bool = False,
) -> typing.TrajectoryFn:
  """Returns trajectory_fn that comuptes decoded trajectory values.

  Args:
    model: model of a dynamical system used to obtain the trajectory.
    start_with_input: whether the firest decoded step in the output trajectory
      should correspond to last input time or first future output.

  Returns:
    Trajectory function that additionally decodes trajectory values.
  """
  def _trajectory_fn(x, forcing_data, outer_steps, inner_steps=1):

    # configure the model.trajectory function with a decoder on the output.
    trajectory_fn = functools.partial(
        model.trajectory,
        forcing_data=forcing_data,
        post_process_fn=with_forcing(model.decode,
                                     model.forcing_fn, forcing_data),
        start_with_input=start_with_input)
    return trajectory_fn(x, outer_steps, inner_steps)

  return _trajectory_fn


def decoded_trajectory_with_inputs_and_forcing(
    model: DynamicalSystem,
    num_init_frames: int,
    start_with_input: bool = False,
) -> typing.TrajectoryFn:
  """Returns trajectory_fn operating on decoded input and forcing data.

  The returned function uses `num_init_frames` of the physics space trajectory
  provided as an input to model.encode_fn to initialize the model state, then
  unrolls the trajectory of specified length that is decoded to the physics
  space using `model.decode_fn`.

  Args:
    model: model of a dynamical system used to obtain the trajectory.
    num_init_frames: number of time frames used from the physics trajectory to
      initialize the model state.
    start_with_input: whether the firest decoded step in the output trajectory
      should correspond to last input time or first future output.

  Returns:
    Trajectory function that operates on physics space trajectories
    and returns unrolls in physics space.
  """
  def _trajectory_fn(x, forcing_data, outer_steps, inner_steps=1):

    # configure the model.trajectory function with a decoder on the output.
    trajectory_fn = decoded_trajectory_with_forcing(model, start_with_input)
    trajectory_fn = functools.partial(
        trajectory_fn,
        forcing_data=forcing_data,
        outer_steps=outer_steps,
        inner_steps=inner_steps)
    # add preprocessing to encode input to model state.
    trajectory_fn = with_preprocessing(
        trajectory_fn, with_forcing(model.encode,
                                    model.forcing_fn, forcing_data))
    # concatenate input trajectory to output trajectory for easier comparison.
    trajectory_fn = with_input_included(
        trajectory_fn, num_last_input_frames_to_trim=int(start_with_input))
    # make trajectories operate on full examples by splitting the init.
    trajectory_fn = with_split_input(trajectory_fn, num_init_frames)
    return trajectory_fn(x)

  return _trajectory_fn


def process_trajectory(
    input_trajectory: Pytree,
    process_fn: Callable[[Pytree], Pytree],
) -> Pytree:
  """Processes trajectory by applying `process_fn` along time axis."""
  step_fn = lambda c, x: tuple([None, hk.remat(process_fn)(x)])
  _, out = hk.scan(step_fn, None, xs=input_trajectory)
  return out


def _maybe_to_nodal_with_physics_sharding(x, /, coords):
  x = coordinate_systems.maybe_to_nodal(x, coords)
  x = coords.with_physics_sharding(x)
  return x


def _maybe_to_modal_with_physics_sharding(x, /, coords):
  x = coordinate_systems.maybe_to_modal(x, coords)
  x = coords.with_physics_sharding(x)
  return x


def compute_prediction_representations(
    predicted_trajectory: typing.Pytree,
    forcing_data: typing.ForcingData,
    model: DynamicalSystem,
) -> typing.TrajectoryRepresentations:
  """Computes TrajectoryRepresentations for predicted trajectory.

  Args:
    predicted_trajectory: predictions on `model.coords` coordinates.
    forcing_data: forcing data to be used for encode/decode transformations.
    model: model used for conversion between representations.

  Returns:
    `TrajectoryRepresentations` for predictions.
  """
  decode_fn = with_forcing(model.decode, model.forcing_fn, forcing_data)
  data_to_nodal = functools.partial(
      _maybe_to_nodal_with_physics_sharding, coords=model.output_coords)
  data_to_modal = functools.partial(
      _maybe_to_modal_with_physics_sharding, coords=model.output_coords)
  model_to_nodal = functools.partial(
      _maybe_to_nodal_with_physics_sharding, coords=model.coords)
  model_to_modal = functools.partial(
      _maybe_to_modal_with_physics_sharding, coords=model.coords)
  predicted_data_trajectory = process_trajectory(
      predicted_trajectory, decode_fn)
  # Note: we pass original prediction to the decoder, but use dict for outputs.
  if isinstance(predicted_trajectory, typing.ModelState):
    predicted_trajectory = predicted_trajectory.state
  if dataclasses.is_dataclass(predicted_trajectory):
    # Losses operate on dicts: convert struct to dict if needed.
    predicted_trajectory = predicted_trajectory.asdict()
  return typing.TrajectoryRepresentations(
      data_nodal_trajectory=process_trajectory(
          predicted_data_trajectory, data_to_nodal),
      data_modal_trajectory=process_trajectory(
          predicted_data_trajectory, data_to_modal),
      model_nodal_trajectory=process_trajectory(
          predicted_trajectory, model_to_nodal),
      model_modal_trajectory=process_trajectory(
          predicted_trajectory, model_to_modal),
  )


def compute_target_representations(
    target_trajectory: typing.Pytree,
    forcing_data: typing.ForcingData,
    model: DynamicalSystem,
) -> typing.TrajectoryRepresentations:
  """Computes TrajectoryRepresentations for target trajectory.

  Note: currently this method only supports models that use a single time slice
  for initialization.

  Args:
    target_trajectory: target trajectory on `model.output_coords` coordinates.
    forcing_data: forcing data to be used for encode/decode transformations.
    model: model used for conversion between representations.

  Returns:
    `TrajectoryRepresentations` for predictions.
  """
  encode_slice_fn = with_forcing(model.encode, model.forcing_fn, forcing_data)
  encode_fn = lambda tree: encode_slice_fn(  # pylint: disable=g-long-lambda.
      jax.tree_util.tree_map(lambda x: jnp.expand_dims(x, 0), tree))
  data_to_nodal = functools.partial(
      _maybe_to_nodal_with_physics_sharding, coords=model.output_coords)
  data_to_modal = functools.partial(
      _maybe_to_modal_with_physics_sharding, coords=model.output_coords)
  model_to_nodal = functools.partial(
      _maybe_to_nodal_with_physics_sharding, coords=model.coords)
  model_to_modal = functools.partial(
      _maybe_to_modal_with_physics_sharding, coords=model.coords)
  target_model_trajectory = process_trajectory(
      target_trajectory, encode_fn)
  if isinstance(target_model_trajectory, typing.ModelState):
    target_model_trajectory = target_model_trajectory.state
  if dataclasses.is_dataclass(target_model_trajectory):
    # Losses operate on dicts: convert struct to dict if needed.
    target_model_trajectory = target_model_trajectory.asdict()
  return typing.TrajectoryRepresentations(
      data_nodal_trajectory=process_trajectory(
          target_trajectory, data_to_nodal),
      data_modal_trajectory=process_trajectory(
          target_trajectory, data_to_modal),
      model_nodal_trajectory=process_trajectory(
          target_model_trajectory, model_to_nodal),
      model_modal_trajectory=process_trajectory(
          target_model_trajectory, model_to_modal),
  )


def compute_prediction_and_target_representations(
    predicted_model_trajectory: typing.Pytree,
    target_data_trajectory: typing.Pytree,
    forcing_data: typing.ForcingData,
    model: DynamicalSystem,
) -> Tuple[typing.TrajectoryRepresentations, typing.TrajectoryRepresentations]:
  """Computes TrajectoryRepresentations for predicted and target trajectories.

  Note: currently this method only supports models that use a single time slice
  for initialization. While computing all terms seems wasteful, once jit-ed
  all unused computations are optimized away. It is also tempting to compute
  all representations at once, but as of 2023-02-28 compiler doesn't manage to
  remove unused computation from a single primitive.

  Args:
    predicted_model_trajectory: predictions on `model.coords` coordinates.
    target_data_trajectory: target data on `model.output_coords` coordinates.
    forcing_data: forcing data to be used for encode/decode transformations.
    model: model used for conversion between representations.

  Returns:
    Tuple of `TrajectoryRepresentations` for predictions and targets.
  """
  prediction_representations = compute_prediction_representations(
      predicted_model_trajectory, forcing_data, model)
  target_representations = compute_target_representations(
      target_data_trajectory, forcing_data, model)
  return prediction_representations, target_representations


@jax.custom_jvp
def safe_sqrt(x: Array) -> jax.Array:
  """Sqrt(x) with gradient = 0 for x near 0."""
  return jnp.sqrt(x)


@safe_sqrt.defjvp
def safe_sqrt_jvp(
    primals: Array,
    tangents: Array,
) -> tuple[jax.Array, jax.Array]:
  (x,) = primals
  (x_dot,) = tangents
  primal_out = safe_sqrt(x)
  eps = jnp.finfo(x.dtype).eps
  safe_x = jnp.where(x > eps, x, 1.0)
  tangent_out = jnp.where(x > eps, x_dot / (2 * safe_sqrt(safe_x)), 0)
  return primal_out, tangent_out



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines observation specifications for NeuralGCM models."""

from __future__ import annotations

import dataclasses
from typing import Generic, TypeVar

import jax
from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import typing
from penzai.core import struct


F = TypeVar('F', cx.Field, cx.Coordinate)


def _extract_coordinate(field: cx.Field):
  """Returns a single coordinate from a fully labeled field."""
  if field.positional_shape:
    raise ValueError(f'Not all dimensions are labeled {field.dims=}')
  return cx.compose_coordinates(*[field.coords[d] for d in field.dims])


@struct.pytree_dataclass
class TimedObservations(
    Generic[F], typing.ObservationData, typing.ObservationSpecs, struct.Struct
):
  """Specifies data or specs for a collection of observations at a given time.

  This object is a self-dual from ObservationData/ObservationSpecs perspective,
  meaning that it can represent either observation specifications or data,
  depending on the type of `self.fields`.

  Attributes:
    fields: A mapping from field names to their values or coordinates.
    timestamp: A timestamp of the observations.
  """

  fields: dict[str, F] = dataclasses.field(metadata={'pytree_node': True})
  timestamp: jax.Array = dataclasses.field(metadata={'pytree_node': True})

  def get_specs(self) -> TimedObservations:
    if_field = lambda x: isinstance(x, cx.Field)
    to_coord = lambda x: _extract_coordinate(x) if if_field(x) else x
    return jax.tree.map(to_coord, self, is_leaf=if_field)


@struct.pytree_dataclass
class TimedField(
    Generic[F], typing.ObservationData, typing.ObservationSpecs, struct.Struct
):
  """Specifies data or specs for a single observations at a given time.

  This object is a self-dual from ObservationData/ObservationSpecs perspective,
  meaning that it
  can represent either observation specifications or data, depending on the
  type of `self.fields`.

  Attributes:
    field: A field or a coordinate specifying the observation.
    timestamp: A timestamp of the observation.
  """

  field: F = dataclasses.field(metadata={'pytree_node': True})
  timestamp: jax.Array | None = dataclasses.field(
      default=None, metadata={'pytree_node': True}
  )

  def get_specs(self) -> TimedField:
    if_field = lambda x: isinstance(x, cx.Field)
    to_coord = lambda x: _extract_coordinate(x) if if_field(x) else x
    return jax.tree.map(to_coord, self, is_leaf=if_field)
